from glimix_core.mean import OffsetMean
from optimix import Assertion


def test_offsetmean_optimix():
    item0 = 5
    item1 = 10

    a = Assertion(OffsetMean, item0, item1, 0.0, offset=0.5)
    a.assert_layout()
    a.assert_gradient()



from __future__ import division

import collections

from numpy import (
    all as npall,
    asarray,
    atleast_2d,
    clip,
    dot,
    errstate,
    exp,
    full,
    isfinite,
    log,
    maximum,
    sqrt,
    sum as npsum,
    zeros,
)

from glimix_core.util import log2pi
from optimix import Function, Scalar

from ..util import economic_qs_zeros, numbers

try:
    from itertools import izip as zip
except ImportError:
    pass


class MTLMMCore(Function):
    def __init__(self, y, X=None, QS=None, SVD=None):
        Function.__init__(self, logistic=Scalar(0.0))

        y = _check_outcome(y)
        X, n, ntraits = _check_covariates(X, SVD)
        _check_outcome_against_covariates(y, n, ntraits)

        if QS is None:
            QS = economic_qs_zeros(n)
            self.delta = 1.0
            super(MTLMMCore, self).fix("logistic")
        else:
            self.delta = 0.5

        _check_qs(QS, y)

        self.variables().get("logistic").bounds = (-numbers.logmax, +numbers.logmax)
        self._QS = QS
        self._y = y
        self._scale = 1.0
        self._fix_scale = False
        self._fix_beta = False

        self._tM = None
        self._tbeta = None

        self._svd = None
        self._set_X(X=X, SVD=SVD)
        self._verbose = False

    def lml(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            Log of the marginal likelihood.
        """
        if self._verbose:
            print("Scale: {:g}".format(self.scale))
            print("Delta: {:g}".format(self.delta))
            print("Beta: {:}".format(str(self.beta)))
        if self.isfixed("scale"):
            return self._lml_arbitrary_scale()
        return self._lml_optimal_scale()

    def _lml_optimal_scale(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            Log of the marginal likelihood.

        """
        self._update_fixed_effects()

        nm = len(self._y[0]) * len(self._y)
        lml = -nm * log2pi - nm - nm * log(self.scale)
        lml -= sum(npsum(log(D)) for D in self._D)

        return lml / 2

    def _lml_arbitrary_scale(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            Log of the marginal likelihood.
        """
        self._update_fixed_effects()
        s = self.scale

        nm = sum(len(yi) for yi in self._y)
        m = len(self._y)
        lml = -nm * log2pi - nm * log(s)

        lml -= m * sum(npsum(log(D)) for D in self._D)

        d = (
            (i - j for (i, j) in zip(ii, jj)) for (ii, jj) in zip(self._mTQ, self._yTQ)
        )

        tmp = sum(
            sum(dot(j / i, l) for (i, j, l) in zip(self._D, di, yTQ))
            for (di, yTQ) in zip(d, self._yTQ)
        )
        lml += tmp / s

        return lml / 2

    def _optimal_scale(self):
        from numpy_sugar import epsilon

        if not self.isfixed("beta"):
            return self._optimal_scale_using_optimal_beta()

        yTQDiQTy = self._yTQDiQTy
        yTQDiQTm = self._yTQDiQTm
        mTQDiQTm = self._mTQDiQTm
        b = self._tbeta
        p0 = sum(
            sum(i - 2 * dot(j, bi) for (i, j) in zip(ii, jj))
            for (ii, jj, bi) in zip(yTQDiQTy, yTQDiQTm, b)
        )
        p1 = sum(sum(dot(dot(bi, i), bi) for i in ii) for (ii, bi) in zip(mTQDiQTm, b))
        return maximum((p0 + p1) / sum(len(yi) for yi in self._y), epsilon.small)

    def _optimal_scale_using_optimal_beta(self):
        from numpy_sugar import epsilon

        yTQDiQTy = self._yTQDiQTy
        yTQDiQTm = self._yTQDiQTm
        b = self._tbeta
        s = sum(
            sum(i - dot(j, bi) for (i, j) in zip(ii, jj))
            for (ii, jj, bi) in zip(yTQDiQTy, yTQDiQTm, b)
        )
        return maximum(s / sum(len(yi) for yi in self._y), epsilon.small)

    def _set_X(self, X=None, SVD=None):
        from numpy_sugar.linalg import ddot, economic_svd

        if SVD is None:
            self._svd = [economic_svd(Xi) for Xi in X]
        else:
            self._svd = SVD
        self._tM = [ddot(svd[0], sqrt(svd[1])) for svd in self._svd]
        self._tbeta = [zeros(tM.shape[1]) for tM in self._tM]

    def _update_fixed_effects(self):
        from numpy_sugar.linalg import rsolve

        if self.isfixed("beta"):
            return

        for (i, j, bi) in zip(self._yTQDiQTm, self._mTQDiQTm, self._tbeta):
            i = list(i)
            j = list(j)
            nominator = i[0]
            denominator = j[0]

            if len(i) > 1:
                nominator += i[1]
                denominator += j[1]

            bi[:] = rsolve(denominator, nominator)

    @property
    def _D(self):
        D = []
        n = self._y[0].shape[0]
        if self._QS[1].size > 0:
            D += [self._QS[1] * (1 - self.delta) + self.delta]
        if self._QS[1].size < n:
            D += [full(n - self._QS[1].size, self.delta)]
        return D

    @property
    def _mTQDiQTm(self):
        return ((dot(i / j, i.T) for (i, j) in zip(ii, self._D)) for ii in self._tMTQ)

    @property
    def _mTQ(self):
        return ((dot(m.T, Q) for Q in self._QS[0] if Q.size > 0) for m in self.mean)

    @property
    def _tMTQ(self):
        return ((tM.T.dot(Q) for Q in self._QS[0] if Q.size > 0) for tM in self._tM)

    @property
    def _yTQ(self):
        return ((dot(yi.T, Q) for Q in self._QS[0] if Q.size > 0) for yi in self._y)

    @property
    def _yTQQTy(self):
        return ((yTQi ** 2 for yTQi in yTQ) for yTQ in self._yTQ)

    @property
    def _yTQDiQTy(self):
        return ((npsum(i / j) for (i, j) in zip(ii, self._D)) for ii in self._yTQQTy)

    @property
    def _yTQDiQTm(self):
        yTQ = self._yTQ
        D = self._D
        tMTQ = self._tMTQ
        return (
            (dot(i / j, l.T) for (i, j, l) in zip(ii, D, ll))
            for (ii, ll) in zip(yTQ, tMTQ)
        )

    @property
    def X(self):
        r"""Covariates set by the user.

        It has to be a matrix of number-of-samples by number-of-covariates.

        Returns
        -------
        :class:`numpy.ndarray`
            Covariates.
        """
        from numpy_sugar.linalg import ddot

        return dot(self._svd[0], ddot(self._svd[1], self._svd[2], left=True))

    @X.setter
    def X(self, X):
        self._set_X(X)

    @property
    def beta(self):
        r"""Fixed-effect sizes.

        The optimal fixed-effect sizes is given by any solution to equation

        .. math::

            (\mathrm Q^{\intercal}\mathrm X)^{\intercal}
                \mathrm D^{-1}
                (\mathrm Q^{\intercal}\mathrm X)
                \boldsymbol\beta =
                (\mathrm Q^{\intercal}\mathrm X)^{\intercal}
                \mathrm D^{-1}
                (\mathrm Q^{\intercal}\mathbf y).

        Returns
        -------
        :class:`numpy.ndarray`
            Optimal fixed-effect sizes.
        """
        from numpy_sugar.linalg import ddot, rsolve

        SVs = (ddot(svd[0], sqrt(svd[1])) for svd in self._svd)
        z = (rsolve(SVsi, m) for (SVsi, m) in zip(SVs, self.mean))
        VsD = (ddot(sqrt(svd[1]), svd[2]) for svd in self._svd)
        return [rsolve(VsDi, zi) for (VsDi, zi) in zip(VsD, z)]

    @beta.setter
    def beta(self, beta):
        from numpy_sugar.linalg import ddot

        beta = (asarray(b, float) for b in beta)
        VsD = (ddot(sqrt(svd[1]), svd[2]) for svd in self._svd)
        for (tb, VsDi, b) in zip(self._tbeta, VsD, beta):
            tb[:] = dot(VsDi, b)

    @property
    def delta(self):
        r"""Variance ratio between ``K`` and ``I``."""
        from numpy_sugar import epsilon

        v = float(self.variables().get("logistic").value)
        with errstate(over="ignore", under="ignore"):
            v = 1 / (1 + exp(-v))
        return clip(v, epsilon.tiny, 1 - epsilon.tiny)

    @delta.setter
    def delta(self, delta):
        from numpy_sugar import epsilon

        delta = clip(delta, epsilon.tiny, 1 - epsilon.tiny)
        self.variables().set(dict(logistic=log(delta / (1 - delta))))

    @property
    def mean(self):
        r"""Mean of the prior.

        Formally, :math:`\mathbf m = \mathrm X \boldsymbol\beta`.

        Returns
        -------
        array_like
            Mean of the prior.
        """
        return [dot(tM, tb) for (tM, tb) in zip(self._tM, self._tbeta)]

    @property
    def scale(self):
        r"""Scaling factor.

        Returns
        -------
        float
            Current scale if fixed; optimal scale otherwise.
        """
        if self.isfixed("scale"):
            return self._scale
        return self._optimal_scale()

    @scale.setter
    def scale(self, scale):
        self._scale = scale

    def mean_star(self, Xstar):
        return dot(Xstar, self.beta)

    def variance_star(self, kss):
        return kss * self.v0 + self.v1

    def covariance_star(self, ks):
        return ks * self.v0

    def covariance(self):
        r"""Covariance of the prior.

        Returns
        -------
        :class:`numpy.ndarray`
            :math:`v_0 \mathrm K + v_1 \mathrm I`.
        """
        from numpy_sugar.linalg import ddot, sum2diag

        Q0 = self._QS[0][0]
        S0 = self._QS[1]
        return sum2diag(dot(ddot(Q0, self.v0 * S0), Q0.T), self.v1)


def _check_outcome(y):
    from numpy_sugar import is_all_finite

    if hasattr(y, "ndim"):
        if y.ndim == 1:
            y = [y]
        elif y.ndim == 2:
            y = y.T
        else:
            raise ValueError("Unrecognized number of dimensions of the outcome array.")
    else:
        if isinstance(y, collections.abc.Sequence):
            if not hasattr(y[0], "ndim") and not isinstance(
                y[0], collections.abc.Sequence
            ):
                y = [asarray(y, float)]
        else:
            y = [y]
    y = [asarray(yi, float).ravel() for yi in y]
    if not all(is_all_finite(yi) for yi in y):
        raise ValueError("There are non-finite values in the outcome.")

    if not all([npall(isfinite(yi)) for yi in y]):
        raise ValueError("Not all values are finite in the outcome array.")

    return y


def _check_covariates(X, SVD):
    if X is not None:
        if hasattr(X, "ndim"):
            if X.ndim == 1 or X.ndim == 2:
                X = [X]
            else:
                raise ValueError(
                    "Unrecognized number of dimensions of the covariates array."
                )
        else:
            if isinstance(X, collections.abc.Sequence):
                if not hasattr(X[0], "ndim") and not isinstance(
                    X[0], collections.abc.Sequence
                ):
                    X = [asarray(X, float)]
            else:
                X = [X]

        X = [atleast_2d(asarray(Xi, float).T).T for Xi in X]
        if not all([npall(isfinite(Xi)) for Xi in X]):
            msg = "Not all values are finite in the covariates matrix."
            raise ValueError(msg)
        n = X[0].shape[0]
        ntraits = len(X)
    else:
        n = SVD[0][0].shape[0]
        ntraits = len(SVD)

    return (X, n, ntraits)


def _check_qs(QS, y):
    if not isinstance(QS, tuple):
        raise ValueError("I was expecting a tuple for the covariance ")

    if QS[0][0].shape[0] != y[0].shape[0]:
        raise ValueError(
            "Number of samples differs between outcome"
            " and covariance matrix decomposition."
        )


def _check_outcome_against_covariates(y, n, ntraits):
    if any([yi.shape[0] != n for yi in y]):
        raise ValueError("Number of samples differs between outcome and covariates.")

    if len(y) != ntraits:
        raise ValueError("Number of traits differs between outcome and covariates.")



from numpy import abs as npy_abs, array, errstate, maximum, nan_to_num, sqrt


def _norm(x0, x1):
    m = maximum(npy_abs(x0), npy_abs(x1))
    with errstate(invalid="ignore"):
        a = (x0 / m) * (x0 / m)
        b = (x1 / m) * (x1 / m)
        return nan_to_num(m * sqrt(a + b))


def hsolve(A00, A01, A11, y0, y1):
    r"""Solver for the linear equations of two variables and equations only.

    It uses Householder reductions to solve ``Ax = y`` in a robust manner.

    Parameters
    ----------
    A : array_like
        Coefficient matrix.
    y : array_like
        Ordinate values.

    Returns
    -------
    :class:`numpy.ndarray`  Solution ``x``.
    """
    from numpy_sugar import epsilon

    n = _norm(A00, A01)
    u0 = A00 - n
    u1 = A01
    nu = _norm(u0, u1)

    with errstate(invalid="ignore", divide="ignore"):
        v0 = nan_to_num(u0 / nu)
        v1 = nan_to_num(u1 / nu)

    B00 = 1 - 2 * v0 * v0
    B01 = 0 - 2 * v0 * v1
    B11 = 1 - 2 * v1 * v1

    D00 = B00 * A00 + B01 * A01
    D01 = B00 * A01 + B01 * A11
    D11 = B01 * A01 + B11 * A11

    b0 = y0 - 2 * y0 * v0 * v0 - 2 * y1 * v0 * v1
    b1 = y1 - 2 * y0 * v1 * v0 - 2 * y1 * v1 * v1

    n = _norm(D00, D01)
    u0 = D00 - n
    u1 = D01
    nu = _norm(u0, u1)

    with errstate(invalid="ignore", divide="ignore"):
        v0 = nan_to_num(u0 / nu)
        v1 = nan_to_num(u1 / nu)

    E00 = 1 - 2 * v0 * v0
    E01 = 0 - 2 * v0 * v1
    E11 = 1 - 2 * v1 * v1

    F00 = E00 * D00 + E01 * D01
    F01 = E01 * D11
    F11 = E11 * D11

    F11 = (npy_abs(F11) > epsilon.small) * F11

    with errstate(divide="ignore", invalid="ignore"):
        Fi00 = nan_to_num(F00 / F00 / F00)
        Fi11 = nan_to_num(F11 / F11 / F11)
        Fi10 = nan_to_num(-(F01 / F00) * Fi11)

    c0 = Fi00 * b0
    c1 = Fi10 * b0 + Fi11 * b1

    x0 = E00 * c0 + E01 * c1
    x1 = E01 * c0 + E11 * c1

    return array([x0, x1])



from __future__ import unicode_literals

from numpy.random import RandomState
from numpy.testing import assert_, assert_allclose

from glimix_core.lik import (
    BernoulliProdLik,
    BinomialProdLik,
    DeltaProdLik,
    PoissonProdLik,
)
from glimix_core.link import ProbitLink


def test_delta_prod_lik():
    random = RandomState(0)

    lik = DeltaProdLik(ProbitLink())

    assert_(lik.name == "Delta")

    lik.outcome = [1, 0, 1]
    assert_allclose(lik.outcome, [1, 0, 1])

    assert_(lik.sample_size == 3)

    assert_allclose(lik.mean([-1, 0, 0.5]), [-1, 0, 0.5])
    assert_allclose(lik.sample([-10, 0, 0.5], random), [-10, 0, 0.5])


def test_bernoulli_prod_lik():
    random = RandomState(0)

    lik = BernoulliProdLik(ProbitLink())

    assert_(lik.name == "Bernoulli")

    lik.outcome = [1, 0, 1]
    assert_allclose(lik.outcome, [1, 0, 1])

    assert_(lik.sample_size == 3)
    assert_allclose(lik.mean([-1, 0, 0.5]), [0.15865525, 0.5, 0.69146246])
    assert_allclose(lik.sample([-10, 0, 0.5], random), [0, 1, 1])


def test_binomial_prod_lik():
    random = RandomState(0)

    lik = BinomialProdLik([6, 2, 3], ProbitLink())
    assert_allclose(lik.ntrials, [6, 2, 3])

    assert_(lik.name == "Binomial")

    lik.nsuccesses = [4, 0, 1]
    assert_allclose(lik.nsuccesses, [4, 0, 1])

    assert_(lik.sample_size == 3)
    assert_allclose(lik.mean([-1, 0, 0.5]), [0.15865525, 0.5, 0.69146246])
    assert_allclose(lik.sample([-10, 0, 0.5], random), [0, 1, 2])


def test_poisson_prod_lik():
    random = RandomState(0)

    lik = PoissonProdLik(ProbitLink())

    assert_(lik.name == "Poisson")

    lik.noccurrences = [1, 4, 3]
    assert_allclose(lik.noccurrences, [1, 4, 3])

    assert_(lik.sample_size == 3)
    assert_allclose(lik.mean([-1, 0, 0.5]), [0.15865525, 0.5, 0.69146246])
    assert_allclose(lik.sample([-10, 0, 0.5], random), [0, 1, 1])

    lik = PoissonProdLik()

    assert_(lik.name == "Poisson")

    lik.noccurrences = [1, 4, 3]
    assert_allclose(lik.noccurrences, [1, 4, 3])

    assert_(lik.sample_size == 3)
    assert_allclose(
        lik.mean([-1, 0, 0.5]), [0.36787944117144233, 1.0, 1.6487212707001282]
    )
    assert_allclose(lik.sample([-10, 0, 0.5], random), [0, 3, 4])



from __future__ import absolute_import, division, unicode_literals

from copy import deepcopy

from numpy import dot, exp, eye, log, pi, trace, zeros
from numpy.linalg import slogdet, solve

from ..lmm import FastScanner
from ._glmm import GLMM


class GLMMNormal(GLMM):
    r"""LMM with heterogeneous Normal likelihoods.

    We model

    .. math::

        \tilde{\boldsymbol\mu} \sim \mathcal N(\mathrm X\boldsymbol\beta,
        v_0 \mathrm K + v_1 \mathrm I + \tilde{\Sigma}),

    where :math:`\tilde{\boldsymbol\mu}` and :math:`\tilde{\Sigma}` are
    typically given by EP approximation to non-Normal likelihood (please, refer
    to :class:`glimix_core.expfam.GLMMExpFam`).
    If that is the case, :math:`\tilde{\Sigma}` is a diagonal matrix with non-negative
    values.
    Those EP parameters are given to this class via their natural forms:
    ``eta`` and ``tau``.

    Parameters
    ----------
    eta : array_like
        :math:`[\tilde{\mu}_0\tilde{\sigma}^{-2}_0 \quad \tilde{\mu}_1\tilde{\sigma}^{-2}_1 \quad\cdots\quad \tilde{\mu}_n\tilde{\sigma}^{-2}_n]`.
    tau : array_like
        :math:`[\tilde{\sigma}^{-2}_0\quad\tilde{\sigma}^{-2}_1 \quad\cdots\quad \tilde{\sigma}^{-2}_n]`.
    X : array_like
        Covariates.
    QS : tuple
        Economic eigendecomposition of :math:`\mathrm K`.
    """

    def __init__(self, eta, tau, X, QS=None):
        self._cache = dict(value=None, grad=None)
        GLMM.__init__(self, eta, ("normal", tau), X, QS)
        self.variables().get("beta").listen(self.clear_cache)
        self.variables().get("logscale").listen(self.clear_cache)
        self.variables().get("logitdelta").listen(self.clear_cache)

    def __copy__(self):
        gef = GLMMNormal(self.eta, self.tau, self._X, self._QS)
        GLMM._copy_to(self, gef)
        gef.__dict__["_cache"] = deepcopy(self._cache)
        return gef

    def clear_cache(self, _=None):
        self._cache["value"] = None
        self._cache["grad"] = None

    @property
    def beta(self):
        return GLMM.beta.fget(self)

    @beta.setter
    def beta(self, v):
        GLMM.beta.fset(self, v)
        self.clear_cache()

    @property
    def logitdelta(self):
        return GLMM.logitdelta.fget(self)

    @logitdelta.setter
    def logitdelta(self, v):
        GLMM.logitdelta.fset(self, v)
        self.clear_cache()

    @property
    def logscale(self):
        return GLMM.logscale.fget(self)

    @logscale.setter
    def logscale(self, v):
        GLMM.logscale.fset(self, v)
        self.clear_cache()

    def fix(self, var_name):
        GLMM.fix(self, var_name)
        self.clear_cache()

    @property
    def eta(self):
        return self._y

    @property
    def tau(self):
        return self._lik[1]

    def get_fast_scanner(self):
        r"""Return :class:`glimix_core.lmm.FastScanner` for the current
        delta."""
        from numpy_sugar.linalg import ddot, economic_qs, sum2diag

        y = self.eta / self.tau

        if self._QS is None:
            K = eye(y.shape[0]) / self.tau
        else:
            Q0 = self._QS[0][0]
            S0 = self._QS[1]
            K = dot(ddot(Q0, self.v0 * S0), Q0.T)
            K = sum2diag(K, 1 / self.tau)

        return FastScanner(y, self._X, economic_qs(K), self.v1)

    def gradient(self):
        from numpy_sugar.linalg import ddot, sum2diag

        if self._cache["grad"] is not None:
            return self._cache["grad"]

        scale = exp(self.logscale)
        delta = 1 / (1 + exp(-self.logitdelta))

        v0 = scale * (1 - delta)
        v1 = scale * delta

        mu = self.eta / self.tau
        n = len(mu)
        if self._QS is None:
            K = zeros((n, n))
        else:
            Q0 = self._QS[0][0]
            S0 = self._QS[1]
            K = dot(ddot(Q0, S0), Q0.T)

        A = sum2diag(sum2diag(v0 * K, v1), 1 / self.tau)
        X = self._X

        m = mu - self.mean()
        g = dict()
        Aim = solve(A, m)

        g["beta"] = dot(m, solve(A, X))

        Kd0 = sum2diag((1 - delta) * K, delta)
        Kd1 = sum2diag(-scale * K, scale)

        g["scale"] = -trace(solve(A, Kd0))
        g["scale"] += dot(Aim, dot(Kd0, Aim))
        g["scale"] *= 1 / 2

        g["delta"] = -trace(solve(A, Kd1))
        g["delta"] += dot(Aim, dot(Kd1, Aim))
        g["delta"] *= 1 / 2

        ed = exp(-self.logitdelta)
        es = exp(self.logscale)

        grad = dict()
        grad["logitdelta"] = g["delta"] * (ed / (1 + ed)) / (1 + ed)
        grad["logscale"] = g["scale"] * es
        grad["beta"] = g["beta"]

        self._cache["grad"] = grad

        return self._cache["grad"]

    def set_variable_bounds(self, var_name, bounds):
        GLMM.set_variable_bounds(self, var_name, bounds)
        self.clear_cache()

    def unfix(self, var_name):
        GLMM.unfix(self, var_name)
        self.clear_cache()

    def value(self):
        r"""Log of the marginal likelihood.

        Formally,

        .. math::

            - \frac{n}{2}\log{2\pi} - \frac{1}{2} \log{\left|
                v_0 \mathrm K + v_1 \mathrm I + \tilde{\Sigma} \right|}
                    - \frac{1}{2}
                    \left(\tilde{\boldsymbol\mu} -
                    \mathrm X\boldsymbol\beta\right)^{\intercal}
                    \left( v_0 \mathrm K + v_1 \mathrm I +
                    \tilde{\Sigma} \right)^{-1}
                    \left(\tilde{\boldsymbol\mu} -
                    \mathrm X\boldsymbol\beta\right)

        Returns
        -------
        float
            :math:`\log{p(\tilde{\boldsymbol\mu})}`
        """
        from numpy_sugar.linalg import ddot, sum2diag

        if self._cache["value"] is not None:
            return self._cache["value"]

        scale = exp(self.logscale)
        delta = 1 / (1 + exp(-self.logitdelta))

        v0 = scale * (1 - delta)
        v1 = scale * delta

        mu = self.eta / self.tau
        n = len(mu)
        if self._QS is None:
            K = zeros((n, n))
        else:
            Q0 = self._QS[0][0]
            S0 = self._QS[1]
            K = dot(ddot(Q0, S0), Q0.T)

        A = sum2diag(sum2diag(v0 * K, v1), 1 / self.tau)
        m = mu - self.mean()

        v = -n * log(2 * pi)
        v -= slogdet(A)[1]
        v -= dot(m, solve(A, m))

        self._cache["value"] = v / 2

        return self._cache["value"]



import numpy as np
from numpy.testing import assert_allclose

from glimix_core.cov import EyeCov
from glimix_core.ggp import ExpFamGP
from glimix_core.lik import BinomialProdLik
from glimix_core.link import LogitLink
from glimix_core.mean import OffsetMean
from glimix_core.random import GGPSampler


def test_ggp_binomial_tobi():
    random = np.random.RandomState(2)

    n = 30

    ntrials = random.randint(30, size=n)
    K = random.randn(n, n)
    K = np.matmul(K, K.T)

    lik = BinomialProdLik(ntrials=ntrials, link=LogitLink())

    mean = OffsetMean()
    mean.set_data(np.ones(n), purpose="sample")
    mean.set_data(np.ones(n), purpose="learn")

    cov2 = EyeCov()
    cov2.set_data((np.arange(n), np.arange(n)), purpose="sample")
    cov2.set_data((np.arange(n), np.arange(n)))

    y = GGPSampler(lik, mean, cov2).sample(random)

    mean.value(np.ones(n))

    ggp = ExpFamGP(y, ("binomial", ntrials), mean, cov2)
    assert_allclose(ggp.lml(), -67.84095700542488)

    ggp.fit(verbose=False)
    assert_allclose(ggp.lml(), -64.26701904994792)



class NamedClass(object):
    def __init__(self):
        self._name = None

    @property
    def name(self):
        r"""Get or set its name.

        Makes printting more user-friendly.
        """
        return self._name

    @name.setter
    def name(self, name):
        self._name = name



from __future__ import division

from numpy import asarray, atleast_1d, exp, log, newaxis

from optimix import Function, Scalar

from ..util.classes import NamedClass


class EyeCov(NamedClass, Function):
    r"""Identity covariance function.

    The mathematical representation is

    .. math::

        f(\mathrm x_0, \mathrm x_1) = s \delta[\mathrm x_0 = \mathrm x_1]

    where :math:`s` is the scale parameter and :math:`\delta` is the Kronecker
    delta.

    Example
    -------

    .. doctest::

        >>> from glimix_core.cov import EyeCov
        >>>
        >>> cov = EyeCov()
        >>> cov.scale = 2.5
        >>>
        >>> item0 = 0
        >>> item1 = 1
        >>> print(cov.value(item0, item1))
        0.0
        >>> g = cov.gradient(item0, item1)
        >>> print(g['logscale'])
        0.0
        >>> item0 = [0, 1, 2]
        >>> item1 = [0, 1, 2]
        >>> print(cov.value(item0, item1))
        [[2.5 0.  0. ]
         [0.  2.5 0. ]
         [0.  0.  2.5]]
        >>> g = cov.gradient(item0, item1)
        >>> print(g['logscale'])
        [[2.5 0.  0. ]
         [0.  2.5 0. ]
         [0.  0.  2.5]]
        >>> print(cov)
        EyeCov()
          scale: 2.5
        >>> cov.name = "identity"
        >>> print(cov)
        EyeCov(): identity
          scale: 2.5
    """

    def __init__(self):
        Function.__init__(self, logscale=Scalar(0.0))
        self.variables().get("logscale").bounds = (-20.0, +10)
        NamedClass.__init__(self)

    @property
    def scale(self):
        r"""Scale parameter."""
        return exp(self.variables().get("logscale").value)

    @scale.setter
    def scale(self, scale):
        from numpy_sugar import epsilon

        scale = max(scale, epsilon.tiny)
        self.variables().get("logscale").value = log(scale)

    def value(self, x0, x1):
        r"""Covariance function evaluated at `(x0, x1)`.

        Parameters
        ----------
        x0 : array_like
            Left-hand side sample or samples.
        x1 : array_like
            Right-hand side sample or samples.

        Returns
        -------
        array_like
            :math:`s \delta[\mathrm x_0 = \mathrm x_1]`.
        """
        x0 = asarray(x0)
        x1 = asarray(x1)
        x0_ = atleast_1d(x0).ravel()[:, newaxis]
        x1_ = atleast_1d(x1).ravel()[newaxis, :]
        v = self.scale * (x0_ == x1_)
        return v.reshape(x0.shape + x1.shape)

    def gradient(self, x0, x1):
        r"""Derivative of the covariance function evaluated at `(x0, x1)`.

        Derivative of the covariance function over :math:`\log(s)`.

        Parameters
        ----------
        x0 : array_like
            Left-hand side sample or samples.
        x1 : array_like
            Right-hand side sample or samples.

        Returns
        -------
        dict
            Dictionary having the `logscale` key for
            :math:`s \delta[\mathrm x_0 = \mathrm x_1]`.
        """
        return dict(logscale=self.value(x0, x1))

    def __str__(self):
        tname = type(self).__name__
        msg = "{}()".format(tname)
        if self.name is not None:
            msg += ": {}".format(self.name)
        msg += "\n"
        msg += "  scale: {}".format(self.scale)
        return msg



from __future__ import division

from numpy import asarray, full, ones

from optimix import Function, Scalar

from ..util.classes import NamedClass


class OffsetMean(NamedClass, Function):
    r"""Offset mean function.

    The mathematical representation is

    .. math::

        f(n) = o \mathbf 1

    where :math:`\mathbf 1` is a :math:`n`-sized vector of ones.

    Example
    -------

    .. doctest::

        >>> from glimix_core.mean import OffsetMean
        >>>
        >>> mean = OffsetMean()
        >>> mean.offset = 2.0
        >>> x = [0, 1, 2]
        >>> print(mean.value(x))
        [2. 2. 2.]
        >>> print(mean.gradient(x))
        {'offset': array([1., 1., 1.])}
        >>> print(mean)
        OffsetMean()
          offset: 2.0
        >>> mean.name = "mean-name"
        >>> print(mean)
        OffsetMean(): mean-name
          offset: 2.0
    """

    def __init__(self):
        Function.__init__(self, offset=Scalar(0.0))
        self.variables().get("offset").bounds = (-200.0, +200)
        NamedClass.__init__(self)

    def value(self, x):
        r"""Offset function evaluated at ``x``.

        Parameters
        ----------
        x : array_like
            Sample ids.

        Returns
        -------
        float
            :math:`o \mathbf 1`.
        """
        x = asarray(x)
        return full(x.shape, self.variables().get("offset").value)

    def gradient(self, x):
        r"""Offset function gradient.

        Parameters
        ----------
        x : array_like
            Sample ids.

        Returns
        -------
        dict
            Dictionary having the `offset` key for :math:`\mathbf 1`.
        """
        x = asarray(x)
        return dict(offset=ones(x.shape))

    @property
    def offset(self):
        r"""Offset parameter."""
        return self.variables().get("offset").value

    @offset.setter
    def offset(self, v):
        self.variables().get("offset").value = v

    def __str__(self):
        tname = type(self).__name__
        msg = "{}()".format(tname)
        if self.name is not None:
            msg += ": {}".format(self.name)
        msg += "\n"
        msg += "  offset: {}".format(self.offset)
        return msg



from __future__ import division

import pytest
from numpy import array, inf, nan, ones
from numpy.testing import assert_allclose

from glimix_core.util import check_covariates, check_economic_qs, check_outcome


def test_util_check_economic_qs():

    A = ones((3, 2))
    B = ones((3, 1))
    C = ones(2)

    with pytest.raises(ValueError):
        check_economic_qs(A)

    with pytest.raises(ValueError):
        check_economic_qs((A, C))

    A[0, 0] = inf
    QS = ((A, B), C)

    with pytest.raises(ValueError):
        check_economic_qs(QS)

    A[0, 0] = 1
    C[0] = nan

    with pytest.raises(ValueError):
        check_economic_qs(QS)


def test_util_check_covariates():
    A = ones(2)
    B = ones((1, 2))

    with pytest.raises(ValueError):
        check_covariates(A)

    B[0, 0] = inf
    with pytest.raises(ValueError):
        check_covariates(B)


def test_util_check_outcome():
    y = ones(5)

    y[0] = nan
    with pytest.raises(ValueError):
        check_outcome((y,), "poisson")

    y[0] = 0.5
    want = array([0.5, 1.0, 1.0, 1.0, 1.0])
    assert_allclose(check_outcome(y, "poisson"), want)

    x = ones(4)

    with pytest.raises(ValueError):
        check_outcome((y, x), "bernoulli")

    x = ones(5)

    with pytest.raises(ValueError):
        check_outcome(y, "normal")


def test_util_check_poisson_outcome():
    y = ones(5)
    y[0] = 25000 + 1
    want = array(
        [2.50000000e04, 1.00000000e00, 1.00000000e00, 1.00000000e00, 1.00000000e00]
    )
    with pytest.warns(UserWarning):
        assert_allclose(check_outcome(y, "poisson"), want)



import pytest
from numpy import inf, nan, newaxis
from numpy.random import RandomState
from numpy.testing import assert_allclose

from glimix_core.lmm import MTLMM
from numpy_sugar.linalg import economic_qs_linear


def test_lmm_interface():
    random = RandomState(0)
    y = random.randn(4)
    X = random.randn(5, 2)
    G = random.randn(5, 6)

    QS = economic_qs_linear(G)
    with pytest.raises(ValueError):
        MTLMM(y, X, QS)

    y = random.randn(6)
    X = random.randn(6, 2)
    with pytest.raises(ValueError):
        MTLMM(y, X, QS)

    with pytest.raises(ValueError):
        MTLMM(y, X, X)

    X = random.randn(6)
    G = random.randn(6, 6)
    QS = economic_qs_linear(G)

    lmm = MTLMM(y, X, QS)
    lml0 = lmm.lml()

    lmm = MTLMM(y, X[:, newaxis], QS)
    lml1 = lmm.lml()

    assert_allclose(lml0, -5.88892141334)
    assert_allclose(lml0, lml1)

    y[0] = inf
    with pytest.raises(ValueError):
        MTLMM(y, X[:, newaxis], QS)

    y[0] = 0
    X[3] = nan
    with pytest.raises(ValueError):
        MTLMM(y, X[:, newaxis], QS)


def test_lmm_mt_interface():
    random = RandomState(0)
    y0 = random.randn(4)
    y1 = random.randn(5)
    X = random.randn(5, 2)
    G = random.randn(5, 6)

    QS = economic_qs_linear(G)
    with pytest.raises(ValueError):
        MTLMM([y0, y1], X, QS)

    with pytest.raises(ValueError):
        MTLMM([y1, y0], X, QS)

    y0 = random.randn(5)

    with pytest.raises(ValueError):
        MTLMM([y1, y0], X, QS)

    X0 = random.randn(5, 2)
    X1 = random.randn(5, 3)

    G1 = random.randn(6, 6)
    QS1 = economic_qs_linear(G1)

    with pytest.raises(ValueError):
        MTLMM([y1, y0], [X0, X1], QS1)

    lmm = MTLMM([y0, y1], [X0, X1], QS)
    lml = lmm.lml()

    assert_allclose(lml, -7.713526905947342)

    lmm = MTLMM(y0, X0, QS)
    lml = lmm.lml()
    assert_allclose(lml, -5.721633217309584)

    y0[3] = inf
    with pytest.raises(ValueError):
        MTLMM([y1, y0], [X0, X1], QS1)


def test_lmm_mt_interface_pandas():
    from pandas import Series, DataFrame

    random = RandomState(0)
    y = Series(random.randn(5))
    X = DataFrame(random.randn(5, 2))
    G = random.randn(5, 6)

    QS = economic_qs_linear(G)
    lmm = MTLMM([y, y], [X, X], QS)
    assert_allclose(lmm.lml(), -10.474952610114517)


def test_lmm_interface_pandas():
    from pandas import Series, DataFrame

    random = RandomState(0)
    y = Series(random.randn(5))
    X = DataFrame(random.randn(5, 2))
    G = random.randn(5, 6)

    QS = economic_qs_linear(G)
    lmm = MTLMM(y, X, QS)

    assert_allclose(lmm.lml(), -6.7985689491)



from __future__ import absolute_import, division, unicode_literals

from copy import copy

from numpy import asarray, ascontiguousarray, clip, diag, dot, exp, log, zeros
from numpy.linalg import pinv

from optimix import Function, Scalar, Vector

from ..util import (
    check_covariates,
    check_economic_qs,
    check_outcome,
    economic_qs_zeros,
    normalise_outcome,
    numbers,
)


class GLMM(Function):
    r"""Generalised Linear Gaussian Processes.

    The variances :math:`v_0` and :math:`v_1` are internally replaced by
    the scale and ratio parameters:

    .. math::
        v_0 = s (1 - \delta) ~\text{ and }~
        v_1 = s \delta.

    Let

    .. math::

        \mathrm Q \mathrm E \mathrm Q^{\intercal}
        = \mathrm K = \mathrm G\mathrm G^{\intercal}

    be the eigen decomposition of the random effect's covariance.
    It turns out that the covariance of the prior can be described as

    .. math::

        s \mathrm Q ((1-\delta)\mathrm E
        + \delta\mathrm I) \mathrm Q^{\intercal}.

    This means that :math:`\mathrm Q` does not change during inference, and
    this fact is used in the implementation to speed-up inference.

    Parameters
    ----------
    y : array_like
        Outcome variable.
    lik : tuple
        Likelihood definition. The first item is one of the following likelihood names:
        ``"Bernoulli"``, ``"Binomial"``, ``"Normal"``, and ``"Poisson"``. For
        `Binomial`, the second item is an array of outcomes.
    X : array_like
        Covariates.
    QS : tuple
        Economic eigen decomposition.
    """

    def __init__(self, y, lik, X, QS=None):
        y = ascontiguousarray(normalise_outcome(y), float)
        X = asarray(X, float)

        Function.__init__(
            self,
            beta=Vector(zeros(X.shape[1])),
            logscale=Scalar(0.0),
            logitdelta=Scalar(0.0),
        )

        if not isinstance(lik, (tuple, list)):
            lik = (lik,)

        self._lik = (lik[0].lower(),) + tuple(ascontiguousarray(i) for i in lik[1:])
        self._y = check_outcome(y, self._lik)
        self._X = check_covariates(X)
        if QS is None:
            self._QS = economic_qs_zeros(self._y.shape[0])
        else:
            self._QS = check_economic_qs(QS)
            if self._y.shape[0] != self._QS[0][0].shape[0]:
                raise ValueError("Number of samples in outcome and covariance differ.")

        if self._y.shape[0] != self._X.shape[0]:
            raise ValueError("Number of samples in outcome and covariates differ.")

        self._factr = 1e5
        self._pgtol = 1e-6
        self._verbose = False
        self.set_variable_bounds("logscale", (log(0.001), 6.0))

        self.set_variable_bounds("logitdelta", (-numbers.logmax, +numbers.logmax))

        if lik[0] == "probit":
            self.delta = 0.0
            self.fix("delta")

        self.set_nodata()

    def _copy_to(self, to):
        d = to.variables()
        s = self.variables()

        d.get("beta").value = asarray(s.get("beta").value, float)
        d.get("beta").bounds = s.get("beta").bounds

        for v in ["logscale", "logitdelta"]:
            d.get(v).value = float(s.get(v).value)
            d.get(v).bounds = s.get(v).bounds

    @property
    def beta(self):
        r"""Fixed-effect sizes.

        Returns
        -------
        :class:`numpy.ndarray`
            :math:`\boldsymbol\beta`.
        """
        return asarray(self.variables().get("beta").value, float)

    @beta.setter
    def beta(self, v):
        self.variables().get("beta").value = v

    def copy(self):
        r"""Create a copy of this object."""
        return copy(self)

    def covariance(self):
        r"""Covariance of the prior.

        Returns
        -------
        :class:`numpy.ndarray`
            :math:`v_0 \mathrm K + v_1 \mathrm I`.
        """
        from numpy_sugar.linalg import ddot, sum2diag

        Q0 = self._QS[0][0]
        S0 = self._QS[1]
        return sum2diag(dot(ddot(Q0, self.v0 * S0), Q0.T), self.v1)

    @property
    def delta(self):
        r"""Get or set the ratio of variance between ``K`` and ``I``.

        Returns
        -------
        float
            :math:`\delta`.
        """
        return 1 / (1 + exp(-self.logitdelta))

    @delta.setter
    def delta(self, v):
        from numpy_sugar import epsilon

        v = clip(v, epsilon.small, 1 - epsilon.small)
        self.logitdelta = log(v / (1 - v))

    def fix(self, var_name):
        r"""Prevent a variable to be adjusted.

        Parameters
        ----------
        var_name : str
            Variable name.
        """
        Function.fix(self, _to_internal_name(var_name))

    def fit(self, verbose=True, factr=1e5, pgtol=1e-7):
        r"""Maximise the marginal likelihood.

        Parameters
        ----------
        verbose : bool
            ``True`` for progress output; ``False`` otherwise.
            Defaults to ``True``.
        factr : float, optional
            The iteration stops when
            ``(f^k - f^{k+1})/max{|f^k|,|f^{k+1}|,1} <= factr * eps``, where ``eps`` is
            the machine precision.
        pgtol : float, optional
            The iteration will stop when ``max{|proj g_i | i = 1, ..., n} <= pgtol``
            where ``pg_i`` is the i-th component of the projected gradient.

        Notes
        -----
        Please, refer to :func:`scipy.optimize.fmin_l_bfgs_b` for further information
        about ``factr`` and ``pgtol``.
        """
        f = self.feed()
        self._verbose = verbose
        f.maximize(verbose=verbose, factr=factr, pgtol=pgtol)
        self._verbose = False

    def lml(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            :math:`\log p(\mathbf y)`
        """
        return self.value()

    @property
    def logitdelta(self):
        return float(self.variables().get("logitdelta").value)

    @logitdelta.setter
    def logitdelta(self, v):
        self.variables().get("logitdelta").value = v

    @property
    def logscale(self):
        return float(self.variables().get("logscale").value)

    @logscale.setter
    def logscale(self, v):
        self.variables().get("logscale").value = v

    def posteriori_mean(self):
        r""" Mean of the estimated posteriori.

        This is also the maximum a posteriori estimation of the latent variable.
        """
        from numpy_sugar.linalg import rsolve

        Sigma = self.posteriori_covariance()
        eta = self._ep._posterior.eta
        return dot(Sigma, eta + rsolve(GLMM.covariance(self), self.mean()))

    def posteriori_covariance(self):
        r""" Covariance of the estimated posteriori."""
        K = GLMM.covariance(self)
        tau = self._ep._posterior.tau
        return pinv(pinv(K) + diag(1 / tau))

    def mean(self):
        r"""Mean of the prior.

        Returns
        -------
        :class:`numpy.ndarray`
            :math:`\mathrm X\boldsymbol\beta`.
        """
        return dot(self._X, self.beta)

    @property
    def scale(self):
        r"""Get or set the overall variance.

        Returns
        -------
        float
            :math:`s`.
        """
        return exp(self.logscale)

    @scale.setter
    def scale(self, v):
        b = self.variables().get("logscale").bounds
        self.logscale = clip(log(v), b[0], b[1])

    def set_variable_bounds(self, var_name, bounds):
        self.variables().get(var_name).bounds = bounds

    def unfix(self, var_name):
        r"""Let a variable be adjusted.

        Parameters
        ----------
        var_name : str
            Variable name.
        """
        Function.unfix(self, _to_internal_name(var_name))

    @property
    def v0(self):
        r"""First variance.

        Returns
        -------
        float
            :math:`v_0 = s (1 - \delta)`
        """
        return self.scale * (1 - self.delta)

    @property
    def v1(self):
        r"""Second variance.

        Returns
        -------
        float
            :math:`v_1 = s \delta`
        """
        return self.scale * self.delta

    def value(self, *_):
        raise NotImplementedError

    def gradient(self, *_):
        raise NotImplementedError

    def mean_star(self, Xstar):
        return dot(Xstar, self.beta)

    def variance_star(self, kss):
        return kss * self.v0 + self.v1

    def covariance_star(self, ks):
        return ks * self.v0


def _to_internal_name(name):
    translation = dict(scale="logscale", delta="logitdelta", beta="beta")
    return translation[name]



from numpy import dot
from numpy.random import RandomState
from numpy.testing import assert_allclose

from glimix_core.cov import GivenCov
from optimix import Assertion


def test_givencov_optimix():
    item0 = 0
    item1 = 1
    K = RandomState(0).randn(5, 5)
    K = dot(K, K.T)
    a = Assertion(lambda: GivenCov(K), item0, item1, 0.0, logscale=0.0)
    a.assert_layout()
    a.assert_gradient()


def test_givencov_interface():
    K = RandomState(0).randn(5, 5)
    K = dot(K, K.T)
    c = GivenCov(K)
    c.scale = 1.5
    assert_allclose(c.scale, 1.5)



from __future__ import division

import pytest
from numpy import arange, concatenate, inf, nan, ones, sqrt, zeros
from numpy.random import RandomState
from numpy.testing import assert_, assert_allclose

from glimix_core.cov import EyeCov, LinearCov, SumCov
from glimix_core.lik import DeltaProdLik
from glimix_core.lmm import MTLMM
from glimix_core.mean import OffsetMean
from glimix_core.random import GGPSampler
from numpy_sugar.linalg import economic_qs_linear


def test_lmm_fix_unfix():
    random = RandomState(9458)
    n = 30
    X = _covariates_sample(random, n, n + 1)

    offset = 1.0

    y = _outcome_sample(random, offset, X)

    QS = economic_qs_linear(X)

    lmm = MTLMM(y, ones((n, 1)), QS)

    assert_(not lmm.isfixed("delta"))
    lmm.fix("delta")
    assert_(lmm.isfixed("delta"))

    assert_(not lmm.isfixed("scale"))
    lmm.fix("scale")
    assert_(lmm.isfixed("scale"))

    lmm.scale = 1.0
    lmm.delta = 0.5

    lmm.fit(verbose=False)

    assert_allclose(lmm.beta[0], 0.7065598068496923)
    assert_allclose(lmm.scale, 1.0)
    assert_allclose(lmm.delta, 0.5)
    assert_allclose(lmm.v0, 0.5)
    assert_allclose(lmm.v1, 0.5)
    assert_allclose(lmm.lml(), -57.56642490856645)

    lmm.unfix("scale")
    lmm.fit(verbose=False)

    assert_allclose(lmm.beta[0], 0.7065598068496923)
    assert_allclose(lmm.v0, 1.060029052117017)
    assert_allclose(lmm.v1, 1.060029052117017)
    assert_allclose(lmm.lml(), -52.037205784544476)

    lmm.unfix("delta")
    lmm.fit(verbose=False)

    assert_allclose(lmm.beta[0], 0.7065598068496922, rtol=1e-5)
    assert_allclose(lmm.v0, 0.5667112269084563, rtol=1e-5)
    assert_allclose(lmm.v1, 1.3679269553495002, rtol=1e-5)
    assert_allclose(lmm.lml(), -51.84396136865774, rtol=1e-5)

    with pytest.raises(ValueError):
        lmm.fix("deltaa")

    with pytest.raises(ValueError):
        lmm.isfixed("deltaa")

    lmm = MTLMM(y, ones((n, 1)), QS)
    lmm.fix("beta")
    lmm.beta = [1.5]
    assert_allclose(lmm.lml(), -59.02992868385325)
    assert_allclose(lmm.beta[0], 1.5)
    lmm.fit(verbose=False)
    assert_allclose(lmm.lml(), -56.56425503034913)
    assert_allclose(lmm.beta[0], 1.5)
    lmm.unfix("beta")
    assert_allclose(lmm.lml(), -52.2963882611983)
    assert_allclose(lmm.beta[0], 0.7065598068496929)
    lmm.fit(verbose=False)
    assert_allclose(lmm.lml(), -51.84396136865775)
    assert_allclose(lmm.beta[0], 0.7065598068496929)
    lmm.fix("beta")
    lmm.beta = lmm.beta[0] + 0.01
    assert_allclose(lmm.lml(), -51.84505787833421)
    lmm.beta = lmm.beta[0] - 2 * 0.01
    assert_allclose(lmm.lml(), -51.84505787833422)


def test_lmm_unique_outcome():
    random = RandomState(9458)
    N = 5
    X = random.randn(N, N + 1)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])

    QS = economic_qs_linear(X)

    lmm = MTLMM(zeros(N), ones((N, 1)), QS)

    lmm.fit(verbose=False)

    assert_allclose(lmm.beta[0], 0, atol=1e-5)
    assert_allclose(lmm.v0, 0, atol=1e-5)
    assert_allclose(lmm.v1, 0, atol=1e-5)


def test_lmm_nonfinite_outcome():
    random = RandomState(9458)
    N = 5
    QS = economic_qs_linear(random.randn(N, N + 1))
    y = zeros(N)

    y[0] = nan
    with pytest.raises(ValueError):
        MTLMM(y, ones((N, 1)), QS)

    y[0] = -inf
    with pytest.raises(ValueError):
        MTLMM(y, ones((N, 1)), QS)

    y[0] = +inf
    with pytest.raises(ValueError):
        MTLMM(y, ones((N, 1)), QS)


def test_lmm_redundant_covariates_fullrank():
    random = RandomState(9458)
    n = 30
    X = _covariates_sample(random, n, n + 1)

    offset = 1.0

    y = _outcome_sample(random, offset, X)

    QS = economic_qs_linear(X)

    lmm = MTLMM(y, ones((n, 1)), QS)
    lmm.fit(verbose=False)

    assert_allclose(lmm.scale, 1.93463817155, rtol=1e-5)
    assert_allclose(lmm.delta, 0.707071227475, rtol=1e-5)
    assert_allclose(lmm.beta, 0.70655980685, rtol=1e-5)

    M = ones((n, 10))
    lmm = MTLMM(y, M, QS)
    lmm.fit(verbose=False)

    assert_allclose(lmm.scale, 1.93463817155, rtol=1e-5)
    assert_allclose(lmm.delta, 0.707071227475, rtol=1e-5)
    assert_allclose(lmm.beta, 0.070655980685, rtol=1e-5)


def test_lmm_redundant_covariates_lowrank():
    random = RandomState(9458)
    n = 30
    X = _covariates_sample(random, n, n - 1)

    offset = 1.0

    y = _outcome_sample(random, offset, X)

    QS = economic_qs_linear(X)

    lmm = MTLMM(y, ones((n, 1)), QS)
    lmm.fit(verbose=False)

    assert_allclose(lmm.scale, 2.97311575698, rtol=1e-5)
    assert_allclose(lmm.delta, 0.693584745932, rtol=1e-5)
    assert_allclose(lmm.beta, 0.932326853301, rtol=1e-5)

    M = ones((n, 10))
    lmm = MTLMM(y, M, QS)
    lmm.fit(verbose=False)

    assert_allclose(lmm.scale, 2.97311575698, rtol=1e-5)
    assert_allclose(lmm.delta, 0.693584745932, rtol=1e-5)
    assert_allclose(lmm.beta, 0.0932326853301, rtol=1e-5)


def test_lmm_mt():
    random = RandomState(9458)
    n = 30
    X0 = _covariates_sample(random, n, 2)
    X1 = _covariates_sample(random, n, 2)
    G = _covariates_sample(random, n, n + 1)

    offset = 0.5

    y0 = _outcome_sample(random, offset, concatenate([X0, G], axis=1))
    y1 = _outcome_sample(random, offset, concatenate([X1, G], axis=1))

    y = [y0, y1]
    X = [X0, X1]
    X = [concatenate([x, ones((n, 1))], axis=1) for x in [X0, X1]]
    QS = economic_qs_linear(G)

    lmm = MTLMM(y, X, QS)

    assert_(not lmm.isfixed("delta"))
    lmm.fix("delta")
    assert_(lmm.isfixed("delta"))

    assert_(not lmm.isfixed("scale"))
    lmm.fix("scale")
    assert_(lmm.isfixed("scale"))

    lmm.scale = 1.0
    lmm.delta = 0.5

    lmm.fit(verbose=False)

    assert_allclose(
        lmm.beta[0], [1.428343850340893, -0.1237931925675461, 0.4402245351179559]
    )
    assert_allclose(
        lmm.beta[1], [-2.5806732580851466, 0.06110533548877167, 0.33064850937519386]
    )
    assert_allclose(lmm.scale, 1.0)
    assert_allclose(lmm.delta, 0.5)
    assert_allclose(lmm.v0, 0.5)
    assert_allclose(lmm.v1, 0.5)
    assert_allclose(lmm.lml(), -137.23576677088965)

    lmm.unfix("scale")
    lmm.fit(verbose=False)

    assert_allclose(
        lmm.beta[0], [1.428343850340893, -0.1237931925675461, 0.4402245351179559]
    )
    assert_allclose(
        lmm.beta[1], [-2.5806732580851466, 0.06110533548877167, 0.33064850937519386]
    )
    assert_allclose(lmm.v0, 1.4275698502922902)
    assert_allclose(lmm.v1, 1.4275698502922902)
    assert_allclose(lmm.lml(), -114.83256708724761)

    lmm.unfix("delta")
    lmm.fit(verbose=False)

    assert_allclose(
        lmm.beta[0],
        [1.4395573454124395, -0.2542920130035301, 0.4402245351179558],
        rtol=1e-5,
    )
    assert_allclose(
        lmm.beta[1],
        [-2.4641710578781626, -0.031027930770324882, 0.3306485093751937],
        rtol=1e-5,
    )

    assert_allclose(lmm.v0, 0.8077295581705093, rtol=1e-5)
    assert_allclose(lmm.v1, 1.8723369293912564, rtol=1e-5)

    assert_allclose(lmm.lml(), -114.06392411125593, rtol=1e-5)

    with pytest.raises(ValueError):
        lmm.fix("deltaa")

    with pytest.raises(ValueError):
        lmm.isfixed("deltaa")

    lmm = MTLMM(y, X, QS)
    lmm.fix("beta")
    lmm.beta = [[1.5, 1.0, -0.1], [1.1, -1.0, -0.2]]

    assert_allclose(lmm.lml(), -146.72783021789496)
    assert_allclose(lmm.beta[0][0], 1.5)
    lmm.fit(verbose=False)
    assert_allclose(lmm.lml(), -141.28412048234352)
    assert_allclose(lmm.beta[0][0], 1.5)
    lmm.unfix("beta")
    assert_allclose(lmm.lml(), -115.5887453500185)
    assert_allclose(lmm.beta[0][0], 1.508274887219204)
    lmm.fit(verbose=False)
    assert_allclose(lmm.lml(), -114.06392411125593)
    assert_allclose(lmm.beta[0][0], 1.4395573454124395)
    lmm.fix("beta")
    lmm.beta = [lmm.beta[0] + 0.01, lmm.beta[1]]
    assert_allclose(lmm.lml(), -114.06519287511233)
    lmm.beta = [lmm.beta[0] - 2 * 1.01, lmm.beta[1]]
    assert_allclose(lmm.lml(), -143.9577716707082)


def test_lmm_iid_prior():
    random = RandomState(9458)
    n = 30
    X = _covariates_sample(random, n, n + 1)

    offset = 1.0

    y = _outcome_sample(random, offset, X)

    lmm = MTLMM(y, ones((n, 1)), None)

    assert_(lmm.isfixed("delta"))

    assert_allclose(lmm.lml(), -52.29638826846387)
    lmm.fit(verbose=False)

    assert_allclose(lmm.beta[0][0], 0.7065598068496929)
    assert_allclose(lmm.scale, 1.9127630509027338)
    assert_allclose(lmm.delta, 0.9999999999999998)
    assert_allclose(lmm.lml(), -52.29638826846388)


def test_lmm_zero_rank_covariates():
    random = RandomState(9458)
    n = 30
    X = _covariates_sample(random, n, n + 1)

    offset = 1.0

    y = _outcome_sample(random, offset, X)

    QS = economic_qs_linear(X)
    lmm = MTLMM(y, zeros((n, 1)), QS)
    lmm.fit(verbose=False)
    assert_allclose(lmm.lml(), -55.774936794143315)
    assert_allclose(lmm.scale, 2.411989807718206)
    assert_allclose(lmm.delta, 0.9999999979388463)
    assert_allclose(lmm.beta, [[0]])


def _covariates_sample(random, n, p):
    X = random.randn(n, p)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])
    return X


def _outcome_sample(random, offset, X):
    n = X.shape[0]
    mean = OffsetMean()
    mean.offset = offset
    mean.set_data(arange(n), purpose="sample")

    cov_left = LinearCov()
    cov_left.scale = 1.5
    cov_left.set_data((X, X), purpose="sample")

    cov_right = EyeCov()
    cov_right.scale = 1.5
    cov_right.set_data((arange(n), arange(n)), purpose="sample")

    cov = SumCov([cov_left, cov_right])

    lik = DeltaProdLik()

    return GGPSampler(lik, mean, cov).sample(random)



from __future__ import division

from numpy import dot, ones, stack, tril_indices_from, zeros, zeros_like

from optimix import Function, Vector

from ..util.classes import NamedClass


class FreeFormCov(NamedClass, Function):
    r"""General semi-definite positive matrix.

    A :math:`d`-by-:math:`d` covariance matrix :math:`\mathrm K` will have
    ``((d + 1) * d) / 2`` parameters defining the lower triangular part of
    the Cholesky matrix ``L``: :math:`\mathrm L\mathrm L^\intercal=\mathrm K`.

    Parameters
    ----------
    size : int
        Dimension :math:`d` of the free-form covariance matrix.

    Example
    -------

    .. doctest::

        >>> from glimix_core.cov import FreeFormCov
        >>>
        >>> cov = FreeFormCov(2)
        >>> print(cov.value([0, 1], [0, 1]))
        [[1. 1.]
         [1. 2.]]
        >>> print(cov.L)
        [[1. 0.]
         [1. 1.]]
        >>> print(cov.Lu)
        [1. 1. 1.]
        >>> g = cov.gradient([0, 1], [0, 1])
        >>> print(g['Lu'].shape)
        (2, 2, 3)
        >>> print(g['Lu'])
        [[[2. 0. 0.]
          [1. 1. 0.]]
        <BLANKLINE>
         [[1. 1. 0.]
          [0. 2. 2.]]]
        >>> cov.Lu[1] = -2
        >>> print(cov.L)
        [[ 1.  0.]
         [-2.  1.]]
        >>> print(cov.value([0, 1], [0, 1]))
        [[ 1. -2.]
         [-2.  5.]]
        >>> print(cov)
        FreeFormCov()
          Lu: [ 1. -2.  1.]
        >>> cov.name = "covname"
        >>> print(cov)
        FreeFormCov(): covname
          Lu: [ 1. -2.  1.]
    """

    def __init__(self, size):
        size = int(size)
        tsize = ((size + 1) * size) // 2
        self._L = zeros((size, size))
        self._tril = tril_indices_from(self._L)
        self._L[self._tril] = 1
        Function.__init__(self, Lu=Vector(ones(tsize)))
        NamedClass.__init__(self)

    @property
    def L(self):
        """Cholesky decomposition of the covariance matrix."""
        self._L[self._tril] = self.variables().get("Lu").value
        return self._L

    @L.setter
    def L(self, value):
        self._L[:] = value
        self.variables().get("Lu").value = self._L[self._tril]

    @property
    def Lu(self):
        """Cholesky decomposition of the covariance matrix in flat form."""
        return self.variables().get("Lu").value

    @Lu.setter
    def Lu(self, value):
        self.variables().get("Lu").value = value

    def value(self, x0, x1):
        r"""Covariance function evaluated at ``(x0, x1)``.

        Parameters
        ----------
        x0 : array_like
            Left-hand side sample indices.
        x1 : array_like
            Right-hand side sample indices.

        Returns
        -------
        array_like
            Submatrix of :math:`\mathrm L\mathrm L^\intercal`, row and
            column-indexed by ``x0`` and ``x1``.
        """
        return dot(self.L, self.L.T)[x0, ...][..., x1]

    def gradient(self, x0, x1):
        r"""Derivative of the covariance function evaluated at ``(x0, x1)``.

        Derivative over the lower triangular part of :math:`\mathrm L`.

        Parameters
        ----------
        x0 : array_like
            Left-hand side sample indices.
        x1 : array_like
            Right-hand side sample indices.

        Returns
        -------
        dict
            Dictionary having the `Lu` key for the lower triangular part
            of the derivative of :math:`\mathrm L\mathrm L^\intercal`, row and
            column-indexed by ``x0`` and ``x1``.
        """
        L = self.L
        Lo = zeros_like(L)
        grad = []
        for ii in range(len(self._tril[0])):
            row = self._tril[0][ii]
            col = self._tril[1][ii]
            Lo[row, col] = 1
            grad.append(dot(Lo, L.T) + dot(L, Lo.T))
            Lo[row, col] = 0

        grad = [g[x0, ...][..., x1] for g in grad]
        return dict(Lu=stack(grad, axis=-1))

    def __str__(self):
        tname = type(self).__name__
        msg = "{}()".format(tname)
        if self.name is not None:
            msg += ": {}".format(self.name)
        msg += "\n"
        msg += "  Lu: {}".format(self.Lu)
        return msg



from __future__ import division

from numpy import arange, sqrt
from numpy.random import RandomState
from numpy.testing import assert_allclose

from glimix_core.cov import EyeCov, LinearCov, SumCov
from glimix_core.ggp import ExpFamGP
from glimix_core.lik import BernoulliProdLik
from glimix_core.link import LogitLink
from glimix_core.mean import OffsetMean
from glimix_core.random import GGPSampler
from optimix import check_grad


def _get_data():
    random = RandomState(0)
    N = 10
    X = random.randn(N, N + 1)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])
    offset = 1.0

    mean = OffsetMean()
    mean.offset = offset
    mean.set_data(arange(N), purpose="sample")
    mean.set_data(arange(N), purpose="learn")

    cov_left = LinearCov()
    cov_left.scale = 1.5
    cov_left.set_data((X, X), purpose="sample")
    cov_left.set_data((X, X), purpose="learn")

    cov_right = EyeCov()
    cov_right.scale = 1.5
    cov_right.set_data((arange(N), arange(N)), purpose="sample")
    cov_right.set_data((arange(N), arange(N)), purpose="learn")

    cov = SumCov([cov_left, cov_right])

    lik = BernoulliProdLik(LogitLink())

    y = GGPSampler(lik, mean, cov).sample(random)

    return dict(
        mean=mean, cov=cov, lik=lik, y=y, cov_left=cov_left, cov_right=cov_right
    )


def test_expfam_ep():
    data = _get_data()
    ep = ExpFamGP((data["y"],), "bernoulli", data["mean"], data["cov"])
    assert_allclose(ep.feed().value(), -5.031838893222976)


def test_expfam_ep_function():
    data = _get_data()
    ep = ExpFamGP((data["y"],), "bernoulli", data["mean"], data["cov"])

    assert_allclose(check_grad(ep.feed()), 0, atol=1e-4)


def test_expfam_ep_optimize():
    data = _get_data()
    ep = ExpFamGP((data["y"],), "bernoulli", data["mean"], data["cov"])
    data["cov_left"].fix("logscale")
    ep.feed().maximize(verbose=False)
    assert_allclose(data["cov_right"].scale, 0.38152460884420397, atol=1e-5)
    assert_allclose(data["mean"].offset, 2.833961965438454, rtol=1e-6)



from numpy import add

from optimix import FunctionReduce

from ..util.classes import NamedClass


class SumMean(NamedClass, FunctionReduce):
    r"""Sum mean function.

    The mathematical representation is

    .. math::

        f(f_0, f_1, \dots) = f_0 + f_1 + \dots

    Parameters
    ----------
    means : list
        List of mean functions.

    Example
    -------

    .. doctest::

        >>> from numpy import arange
        >>> from glimix_core.mean import OffsetMean, LinearMean, SumMean
        >>>
        >>> X = [[5.1, 1.0],
        ...      [2.1, -0.2]]
        >>>
        >>> mean0 = LinearMean(2)
        >>> mean0.set_data((X, ))
        >>> mean0.effsizes = [-1.0, 0.5]
        >>>
        >>> mean1 = OffsetMean()
        >>> mean1.set_data((arange(2), ))
        >>> mean1.offset = 2.0
        >>>
        >>> mean = SumMean([mean0, mean1])
        >>>
        >>> print(mean.feed().value())
        [-2.6 -0.2]
        >>> g = mean.feed().gradient()
        >>> print(g['sum[0].effsizes'])
        [[5.1, 1.0], [2.1, -0.2]]
        >>> print(g['sum[1].offset'])
        [1. 1.]
        >>> print(mean)
        SumMean(means=...)
          LinearMean(size=2)
            effsizes: [-1.   0.5]
          OffsetMean()
            offset: 2.0
        >>> mean0.name = "A"
        >>> mean1.name = "B"
        >>> mean.name = "A+B"
        >>> print(mean)
        SumMean(means=...): A+B
          LinearMean(size=2): A
            effsizes: [-1.   0.5]
          OffsetMean(): B
            offset: 2.0
    """

    def __init__(self, means):
        self._means = [c for c in means]
        FunctionReduce.__init__(self, self._means, "sum")
        NamedClass.__init__(self)

    def value_reduce(self, values):
        r"""Sum mean function evaluated at :math:`(f_0, f_1, \dots)`.

        Parameters
        ----------
        values : dict
            A value for each function involved in the summation.

        Returns
        -------
        dict
            :math:`f_0 + f_1 + \dots`
        """
        return add.reduce(list(values.values()))

    def gradient_reduce(self, values, gradients):
        r"""Sum of mean function derivatives.

        Parameters
        ----------
        values : dict
            Its value is not used in this particular function. We suggest you to simply
            pass ``None``.
        gradients : dict
            A gradient for each function involved in the summation.

        Returns
        -------
        dict
            Dictionary having a key for each parameter of the underlying
            mean functions.
        """
        grad = dict()
        for (gn, gv) in iter(gradients.items()):
            for n, v in iter(gv.items()):
                grad[gn + "." + n] = v
        return grad

    def __str__(self):
        tname = type(self).__name__
        msg = "{}(means=...)".format(tname)
        if self.name is not None:
            msg += ": {}".format(self.name)
        for m in self._means:
            spl = str(m).split("\n")
            msg = msg + "\n" + "\n".join(["  " + s for s in spl])
        return msg



from __future__ import division

from numpy.random import RandomState


class GPSampler(object):
    r"""Sample from a Gaussian Process.

    Outcome modelled via

    .. math::

        \mathbf y \sim \mathcal N(\mathbf z ~|~ \mathbf m, \mathrm K)
        \mathrm d\mathbf z.

    Parameters
    ----------
    mean : function
        Mean function. (Refer to :doc:`mean`.)
    cov : function
        Covariance function. (Refer to :doc:`cov`.)

    Example
    -------

    .. doctest::

        >>> from numpy.random import RandomState
        >>>
        >>> from glimix_core.example import offset_mean
        >>> from glimix_core.example import linear_eye_cov
        >>> from glimix_core.random import GPSampler
        >>>
        >>> random = RandomState(1)
        >>>
        >>> mean = offset_mean()
        >>> cov = linear_eye_cov()
        >>>
        >>> y = GPSampler(mean, cov).sample(random)
        >>> print(y[:5])  # doctest: +FLOAT_CMP
        [-2.42181498  0.50720447 -1.01053967  0.736624    1.64019063]
    """

    def __init__(self, mean, cov):
        self._mean = mean
        self._cov = cov

    def sample(self, random_state=None):
        r"""Sample from the specified distribution.

        Parameters
        ----------
        random_state : random_state
            Set the initial random state.

        Returns
        -------
        numpy.ndarray
            Sample.
        """
        from numpy_sugar import epsilon
        from numpy_sugar.linalg import sum2diag
        from numpy_sugar.random import multivariate_normal

        if random_state is None:
            random_state = RandomState()

        m = self._mean.feed("sample").value()
        K = self._cov.feed("sample").value().copy()

        sum2diag(K, +epsilon.small, out=K)

        return multivariate_normal(m, K, random_state)



import pytest
from numpy import array, inf, nan, ones

from glimix_core.lmm import FastScanner
from numpy_sugar.linalg import economic_qs


def test_scan_interface():
    y = array([-1.0449132, 1.15229426, 0.79595129])
    low_rank_K = array([[5.0, 14.0, 23.0], [14.0, 50.0, 86.0], [23.0, 86.0, 149.0]])
    QS = economic_qs(low_rank_K)
    X = ones((3, 1))

    y[0] = nan
    with pytest.raises(ValueError):
        FastScanner(y, X, QS, 0.5)

    y[0] = inf
    with pytest.raises(ValueError):
        FastScanner(y, X, QS, 0.5)

    y[0] = 1
    X[0, 0] = nan
    with pytest.raises(ValueError):
        FastScanner(y, X, QS, 0.5)

    y[0] = 1
    X[0, 0] = 1
    with pytest.raises(ValueError):
        FastScanner(y, X, QS, -1)

    with pytest.raises(ValueError):
        FastScanner(y, X, QS, nan)



"""Covariance functions.

EyeCov       Indentity covariance function.
FreeFormCov  Free-form covariance based on Cholesky decomposition.
GivenCov     Fixed covariance function with scaling factor.
LinearCov    Dot-product covariance function.
SumCov       Sum of multiple covariance functions.
"""
from .eye import EyeCov
from .free import FreeFormCov
from .given import GivenCov
from .linear import LinearCov
from .sum import SumCov

__all__ = ["EyeCov", "FreeFormCov", "GivenCov", "LinearCov", "SumCov"]



from .check import check_economic_qs, check_covariates, check_outcome
from .hsolve import hsolve
from .eigen import economic_qs_zeros
from . import numbers
from . import classes
from .normalise import normalise_outcome, normalise_covariance, normalise_covariates

log2pi = 1.837877066409345339081937709124758839607238769531250

__all__ = [
    "check_covariates",
    "check_economic_qs",
    "check_outcome",
    "classes",
    "economic_qs_zeros",
    "hsolve",
    "normalise_covariance",
    "normalise_covariates",
    "normalise_outcome",
    "numbers",
]



from numpy import ascontiguousarray, atleast_2d


def normalise_outcome(y):
    return y


def normalise_covariates(M):
    M = ascontiguousarray(M, float)
    M = atleast_2d(M.T).T
    return M


def normalise_covariance(K):
    K = ascontiguousarray(K, float)
    if K.ndim != 2:
        raise ValueError("Covariance matrix must have two dimensions.")
    if K.shape[0] != K.shape[1]:
        raise ValueError("Covariance matrix must be square.")
    return K



from __future__ import division

from numpy import arange, corrcoef, dot, ones, sqrt
from numpy.random import RandomState
from numpy.testing import assert_allclose
from numpy_sugar.linalg import economic_qs_linear

from glimix_core.cov import EyeCov, LinearCov, SumCov
from glimix_core.lik import DeltaProdLik
from glimix_core.lmm import LMM
from glimix_core.mean import OffsetMean
from glimix_core.random import GGPSampler


def _outcome_sample(random, offset, X):
    n = X.shape[0]
    mean = OffsetMean()
    mean.offset = offset
    mean.set_data(arange(n), purpose="sample")

    cov_left = LinearCov()
    cov_left.scale = 1.5
    cov_left.set_data((X, X), purpose="sample")

    cov_right = EyeCov()
    cov_right.scale = 1.5
    cov_right.set_data((arange(n), arange(n)), purpose="sample")

    cov = SumCov([cov_left, cov_right])

    lik = DeltaProdLik()

    return GGPSampler(lik, mean, cov).sample(random)


def _covariates_sample(random, n, p):
    X = random.randn(n, p)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])
    return X


def test_lmm_prediction():
    random = RandomState(9458)
    n = 30
    X = _covariates_sample(random, n, n + 1)

    offset = 1.0

    y = _outcome_sample(random, offset, X)

    QS = economic_qs_linear(X)

    lmm = LMM(y, ones((n, 1)), QS)

    lmm.fit(verbose=False)

    K = dot(X, X.T)
    pm = lmm.predictive_mean(ones((n, 1)), K, K.diagonal())
    assert_allclose(corrcoef(y, pm)[0, 1], 0.8358820971891354)



from __future__ import division, unicode_literals

from numpy import ascontiguousarray

from ..link import IdentityLink, LogitLink, LogLink


def _sample_doc(func):
    func.__doc__ = r"""Sample from the likelihood distribution.

        Parameters
        ----------
        x : array_like
            Array of likelihood parameters.
        random_state : random_state
            Set the initial random state.

        Returns
        -------
        numpy.ndarray
            Sampled outcome.
        """
    return func


class DeltaProdLik(object):
    r"""Represents a product of Kronecker delta likelihoods.

    The product can be written as

    .. math::

        \prod_i \delta[y_i = x_i]

    Parameters
    ----------
    link : link_func
        Link function establishing :math:`g(y_i) = x_i`. Defaults to ``None``, which
        leads to the identity link function.
    """

    def __init__(self, link=None):
        if link is None:
            link = IdentityLink()
        self._link = link
        self._outcome = None

    @property
    def name(self):
        r"""Get the name of this likelihood."""
        return "Delta"

    @property
    def outcome(self):
        r"""Get or set an array of outcomes."""
        return self._outcome

    @outcome.setter
    def outcome(self, v):
        self._outcome = _aca(v)

    def mean(self, x):
        r"""Outcome mean."""
        return x

    @_sample_doc
    def sample(self, x, random_state=None):
        return _aca(x)

    @property
    def sample_size(self):
        r"""Get the number of samples."""
        return len(self.outcome)


class BernoulliProdLik(object):
    r"""Represents a product of Bernoulli likelihoods.

    The product can be written as

    .. math::

        \prod_i p_i^{y_i} (1-p_i)^{1-y_i}

    where :math:`p_i` is the probability of success.

    Parameters
    ----------
    link : link_func
        Link function establishing :math:`g(p_i) = x_i`. Defaults to ``None``, which
        leads to the :class:`glimix_core.link.LogitLink` link function.
    """

    def __init__(self, link=None):
        if link is None:
            link = LogitLink
        self._link = link
        self._outcome = None

    @property
    def name(self):
        r"""Get the name of this likelihood."""
        return "Bernoulli"

    @property
    def outcome(self):
        r"""Get or set an array of outcomes."""
        return self._outcome

    @outcome.setter
    def outcome(self, v):
        self._outcome = _aca(v)

    def mean(self, x):
        r"""Outcome mean."""
        return self._link.inv(x)

    @_sample_doc
    def sample(self, x, random_state=None):
        import scipy.stats as st

        p = self.mean(x)
        return _aca(st.bernoulli(p).rvs(random_state=random_state))

    @property
    def sample_size(self):
        r"""Get the number of samples."""
        return len(self.outcome)


class BinomialProdLik(object):
    r"""Represents a product of Binomial likelihoods.

    The product can be written as

    .. math::

        \prod_i \binom{n_i}{n_i y_i} p_i^{n_i y_i}
        (1-p_i)^{n_i - n_i y_i}

    where :math:`p_i` is the probability of success.

    Parameters
    ----------
    ntrials : array_like
        Array of number of trials.
    link : link_func
        Link function establishing :math:`g(p_i) = x_i`. Defaults to ``None``, which
        leads to the :class:`glimix_core.link.LogitLink` link function.
    """

    def __init__(self, ntrials, link=None):
        if link is None:
            link = LogitLink
        self._link = link
        self._nsuccesses = None
        self._ntrials = _aca(ntrials)

    @property
    def name(self):
        r"""Get the name of this likelihood."""
        return "Binomial"

    @property
    def ntrials(self):
        r"""Get the array of number of trials."""
        return self._ntrials

    @property
    def nsuccesses(self):
        r"""Get or set an array of successful trials."""
        return self._nsuccesses

    @nsuccesses.setter
    def nsuccesses(self, v):
        self._nsuccesses = _aca(v)

    def mean(self, x):
        r"""Mean of the number of successful trials."""
        return self._link.inv(x)

    @_sample_doc
    def sample(self, x, random_state=None):
        import scipy.stats as st

        p = self.mean(x)
        nt = ascontiguousarray(self._ntrials, dtype=int)
        return _aca(st.binom(nt, p).rvs(random_state=random_state))

    @property
    def sample_size(self):
        r"""Get the number of samples."""
        return len(self.nsuccesses)


class PoissonProdLik(object):
    r"""Represents a product of Poisson likelihoods.

    Parameters
    ----------
    link : link_func
        Link function establishing :math:`g(y_i) = x_i`. Defaults to ``None``, which
        leads to the :class:`glimix_core.link.LogitLink` link function.
    """

    def __init__(self, link=None):
        if link is None:
            link = LogLink()
        self._link = link
        self._noccurrences = None

    @property
    def name(self):
        r"""Get the name of this likelihood."""
        return "Poisson"

    @property
    def noccurrences(self):
        r"""Get or set an array of number of occurrences."""
        return self._noccurrences

    @noccurrences.setter
    def noccurrences(self, v):
        self._noccurrences = _aca(v)

    def mean(self, x):
        r"""Mean of the number of occurrences."""
        return self._link.inv(x)

    @_sample_doc
    def sample(self, x, random_state=None):
        import scipy.stats as st

        lam = self.mean(x)
        return st.poisson(mu=lam).rvs(random_state=random_state)

    @property
    def sample_size(self):
        r"""Get the number of samples."""
        return len(self.noccurrences)


def _aca(x):
    return ascontiguousarray(x, dtype=float)



from __future__ import division

from numpy import arange, array, concatenate, newaxis, ones, sqrt
from numpy.random import RandomState
from numpy.testing import assert_allclose

from glimix_core.cov import EyeCov, LinearCov, SumCov
from glimix_core.lik import DeltaProdLik
from glimix_core.lmm import LMM, FastScanner
from glimix_core.mean import OffsetMean
from glimix_core.random import GGPSampler
from numpy_sugar.linalg import economic_qs, economic_qs_linear


def test_scan_fix_unfix():
    random = RandomState(12)
    n = 100
    X = _covariates_sample(random, n, n + 1)
    offset = 0.1

    y = _outcome_sample(random, offset, X)
    QS = economic_qs_linear(X)
    lmm = LMM(y, ones((n, 1)), QS)

    lmm.fit(verbose=False)

    lml0 = lmm.lml()
    assert_allclose(lml0, -193.88684605236722)
    lmm.fix("scale")
    lml1 = lmm.lml()

    assert_allclose(lml0, lml1)

    scale = lmm.scale
    lmm.fit(verbose=False)
    assert_allclose(scale, lmm.scale)

    assert_allclose(lmm.delta, 0.546799614073)
    lmm.delta = 0.5
    assert_allclose(lmm.delta, 0.5)

    assert_allclose(lmm.lml(), -193.947028697)

    assert_allclose(lmm.scale, 3.10605443333)

    lmm.scale = 0.5

    lmm.fit(verbose=False)

    assert_allclose(lmm.scale, 0.5)
    assert_allclose(lmm.delta, 0.775021320328)
    assert_allclose(lmm.lml(), -351.381863666)

    lmm.fix("delta")
    lmm.delta = 0.1

    assert_allclose(lmm.scale, 0.5)
    assert_allclose(lmm.delta, 0.1)
    assert_allclose(lmm.lml(), -615.1757214529657)


def test_scan_fast_scan():
    random = RandomState(9458)
    n = 30
    X = _covariates_sample(random, n, n + 1)
    offset = 1.0

    y = _outcome_sample(random, offset, X)

    QS = economic_qs_linear(X)

    lmm = LMM(y, ones((n, 1)), QS)

    lmm.fit(verbose=False)

    markers = random.randn(n, 2)

    lmm_ = lmm.copy()
    lmm_.X = concatenate([lmm.X, markers[:, 0][:, newaxis]], axis=1)
    lmm_.fix("delta")
    lmm_.fit(verbose=False)
    lml0 = lmm_.lml()

    lmm_ = lmm.copy()
    lmm_.X = concatenate([lmm.X, markers[:, 1][:, newaxis]], axis=1)
    lmm_.fix("delta")
    lmm_.fit(verbose=False)
    lml1 = lmm_.lml()

    fast_scanner = lmm.get_fast_scanner()

    lmls = fast_scanner.fast_scan(markers, verbose=False)[0]
    assert_allclose(lmls, [lml0, lml1])


def test_scan_fastlmm_redundant_candidates():
    random = RandomState(9458)
    n = 10
    X = _covariates_sample(random, n, n + 1)
    offset = 1.0

    y = _outcome_sample(random, offset, X)

    QS = economic_qs_linear(X)

    M = ones((n, 5))
    lmm = LMM(y, M, QS)

    lmm.fit(verbose=False)

    markers = M.copy()

    fast_scanner = lmm.get_fast_scanner()

    lmls = fast_scanner.fast_scan(markers, verbose=False)[0]
    assert_allclose(lmls, [-9.929912871392519] * 5, rtol=1e-5)


def test_scan_fastlmm_set_scale_1covariate():
    random = RandomState(9458)
    n = 10
    X = _covariates_sample(random, n, n + 1)
    offset = 1.0

    y = _outcome_sample(random, offset, X)

    QS = economic_qs_linear(X)

    M = random.randn(n, 1)
    lmm = LMM(y, M, QS)

    lmm.fit(verbose=False)
    assert_allclose(lmm.scale, 5.282731934070453)
    assert_allclose(lmm.delta, 0.7029974630034005)
    assert_allclose(lmm.beta, [0.0599712498212])

    markers = M.copy() + random.randn(n, 1)

    fast_scanner = lmm.get_fast_scanner()

    fast_scanner.set_scale(1.0)

    lmls, effsizes = fast_scanner.fast_scan(markers, verbose=False)

    assert_allclose(lmls, [-21.577703], rtol=1e-6)
    assert_allclose(effsizes, [1.412239], rtol=1e-6)

    fast_scanner.unset_scale()
    lmls, effsizes = fast_scanner.fast_scan(markers, verbose=False)

    assert_allclose(lmls, [-21.509721], rtol=1e-6)
    assert_allclose(effsizes, [1.412239], rtol=1e-6)


def test_scan_fastlmm_set_scale_1covariate_redundant():
    random = RandomState(9458)
    n = 10
    X = _covariates_sample(random, n, n + 1)
    offset = 1.0

    y = _outcome_sample(random, offset, X)

    QS = economic_qs_linear(X)

    M = random.randn(n, 1)
    lmm = LMM(y, M, QS)

    lmm.fit(verbose=False)

    markers = M.copy()

    fast_scanner = lmm.get_fast_scanner()

    fast_scanner.set_scale(1.0)

    lmls, effsizes = fast_scanner.fast_scan(markers, verbose=False)

    assert_allclose(lmls, [-22.357526], rtol=1e-6)
    assert_allclose(effsizes, [0.029986], rtol=1e-6, atol=1e-6)

    fast_scanner.unset_scale()
    lmls, effsizes = fast_scanner.fast_scan(markers, verbose=False)
    assert_allclose(lmls[0], -22.357525517597185, rtol=1e-6)
    assert_allclose(effsizes[0], 0.02998562491058301, rtol=1e-6, atol=1e-6)


def test_scan_fastlmm_set_scale_multicovariates():
    random = RandomState(9458)
    n = 10
    X = _covariates_sample(random, n, n + 1)
    offset = 1.0

    y = _outcome_sample(random, offset, X)

    QS = economic_qs_linear(X)

    M = random.randn(n, 3)
    lmm = LMM(y, M, QS)

    lmm.fit(verbose=False)

    markers = M.copy()

    fast_scanner = lmm.get_fast_scanner()

    fast_scanner.set_scale(1.0)

    lmls, effsizes = fast_scanner.fast_scan(markers, verbose=False)

    want = [-19.318845, -19.318845, -19.318845]
    assert_allclose(lmls, want, rtol=1e-6, atol=1e-6)

    want = [-0.34615, 1.177541, -0.381578]
    assert_allclose(effsizes, want, rtol=1e-6, atol=1e-6)

    fast_scanner.unset_scale()
    lmls, effsizes = fast_scanner.fast_scan(markers, verbose=False)

    want = [-19.318845, -19.318845, -19.318845]
    assert_allclose(lmls, want, rtol=1e-6, atol=1e-6)

    want = [-0.34615, 1.177541, -0.381578]
    assert_allclose(effsizes, want, rtol=1e-6, atol=1e-6)


def test_scan_difficult_settings_offset():
    y = array([-1.0449132, 1.15229426, 0.79595129])
    low_rank_K = array([[5.0, 14.0, 23.0], [14.0, 50.0, 86.0], [23.0, 86.0, 149.0]])
    full_rank_K = array([[6.0, 14.0, 23.0], [14.0, 51.0, 86.0], [23.0, 86.0, 150.0]])
    low_rank_QS = economic_qs(low_rank_K)
    full_rank_QS = economic_qs(full_rank_K)
    X = ones((3, 1))

    M = array(
        [[0.88766985, -1.80940339], [0.00822629, -0.4488265], [0.55807272, -2.00868376]]
    )
    scanner = FastScanner(y, X, low_rank_QS, 0)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [28.896466, 28.896466], rtol=1e-6, atol=1e-6)
    assert_allclose(effsizes, [-5.5851957509459353, -9.2375605785935715])

    X = ones((3, 1))
    M = array(
        [[0.88766985, -1.80940339], [0.00822629, -0.4488265], [0.55807272, -2.00868376]]
    )
    scanner = FastScanner(y, X, low_rank_QS, 0.75)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [0.823505300549, -0.607251148058])
    assert_allclose(effsizes, [-1.79486991632, 0.872439001906])

    M = array([[0.0, -1.80940339], [0.0, -0.4488265], [0.0, -2.00868376]])
    scanner = FastScanner(y, X, low_rank_QS, 0)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [8.704135, 28.896466], rtol=1e-6)
    assert_allclose(effsizes[1], -9.23756057859)

    M = array([[0.0, -1.80940339], [0.0, -0.4488265], [0.0, -2.00868376]])
    scanner = FastScanner(y, X, low_rank_QS, 0.75)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [-6.137441, -0.607251], rtol=1e-6, atol=1e-6)
    assert_allclose(effsizes[1], 0.872439001906)

    M = array(
        [[0.88766985, -1.80940339], [0.00822629, -0.4488265], [0.55807272, -2.00868376]]
    )
    scanner = FastScanner(y, X, full_rank_QS, 0)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [0.636421389625, -0.795311239268])
    assert_allclose(effsizes, [-1.79757295968, 0.871806031702])

    M = array(
        [[0.88766985, -1.80940339], [0.00822629, -0.4488265], [0.55807272, -2.00868376]]
    )
    scanner = FastScanner(y, X, full_rank_QS, 0.75)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [0.251547678451, -1.18305656511])
    assert_allclose(effsizes, [-1.80551427016, 0.86994164292])

    M = array([[0.0, -1.80940339], [0.0, -0.4488265], [0.0, -2.00868376]])
    scanner = FastScanner(y, X, full_rank_QS, 0)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [-5.911136, -0.795311], atol=1e-6, rtol=1e-6)
    assert_allclose(effsizes[1], 0.871806031702)

    M = array([[0.0, -1.80940339], [0.0, -0.4488265], [0.0, -2.00868376]])
    scanner = FastScanner(y, X, full_rank_QS, 0.75)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [-5.509792, -1.183057], atol=1e-6, rtol=1e-6)
    assert_allclose(effsizes[1], 0.86994164292)


def test_scan_difficult_settings_multicovariates():
    y = array([-1.0449132, 1.15229426, 0.79595129])
    low_rank_K = array([[5.0, 14.0, 23.0], [14.0, 50.0, 86.0], [23.0, 86.0, 149.0]])
    full_rank_K = array([[6.0, 14.0, 23.0], [14.0, 51.0, 86.0], [23.0, 86.0, 150.0]])
    low_rank_QS = economic_qs(low_rank_K)
    full_rank_QS = economic_qs(full_rank_K)
    X = array(
        [
            [-0.40592765, 1.04348945],
            [0.92275415, -0.32394197],
            [-0.98197991, 1.22912219],
        ]
    )

    M = array(
        [[0.88766985, -1.80940339], [0.00822629, -0.4488265], [0.55807272, -2.00868376]]
    )
    scanner = FastScanner(y, X, low_rank_QS, 0)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [28.896466, 28.896466], atol=1e-6, rtol=1e-6)
    assert_allclose(effsizes, [-1.195493, 0.318417], atol=1e-6, rtol=1e-6)

    X = ones((3, 1))
    M = array(
        [[0.88766985, -1.80940339], [0.00822629, -0.4488265], [0.55807272, -2.00868376]]
    )
    scanner = FastScanner(y, X, low_rank_QS, 0.75)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [0.823505300549, -0.607251148058])
    assert_allclose(effsizes, [-1.79486991632, 0.872439001906])

    M = array([[0.0, -1.80940339], [0.0, -0.4488265], [0.0, -2.00868376]])
    scanner = FastScanner(y, X, low_rank_QS, 0)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [8.704135, 28.896466])
    assert_allclose(effsizes[1], -9.23756057859)

    M = array([[0.0, -1.80940339], [0.0, -0.4488265], [0.0, -2.00868376]])
    scanner = FastScanner(y, X, low_rank_QS, 0.75)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [-6.137441, -0.607251], atol=1e-6, rtol=1e-6)
    assert_allclose(effsizes[1], 0.872439001906)

    M = array(
        [[0.88766985, -1.80940339], [0.00822629, -0.4488265], [0.55807272, -2.00868376]]
    )
    scanner = FastScanner(y, X, full_rank_QS, 0)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [0.636421389625, -0.795311239268])
    assert_allclose(effsizes, [-1.79757295968, 0.871806031702])

    M = array(
        [[0.88766985, -1.80940339], [0.00822629, -0.4488265], [0.55807272, -2.00868376]]
    )
    scanner = FastScanner(y, X, full_rank_QS, 0.75)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [0.251547678451, -1.18305656511])
    assert_allclose(effsizes, [-1.80551427016, 0.86994164292])

    M = array([[0.0, -1.80940339], [0.0, -0.4488265], [0.0, -2.00868376]])
    scanner = FastScanner(y, X, full_rank_QS, 0)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [-5.911136, -0.795311], atol=1e-6, rtol=1e-6)
    assert_allclose(effsizes[1], 0.871806031702)

    M = array([[0.0, -1.80940339], [0.0, -0.4488265], [0.0, -2.00868376]])
    scanner = FastScanner(y, X, full_rank_QS, 0.75)
    lmls, effsizes = scanner.fast_scan(M, verbose=False)
    assert_allclose(lmls, [-5.509792, -1.183057], atol=1e-6, rtol=1e-6)


def _outcome_sample(random, offset, X):
    n = X.shape[0]
    mean = OffsetMean()
    mean.offset = offset
    mean.set_data(arange(n), purpose="sample")

    cov_left = LinearCov()
    cov_left.scale = 1.5
    cov_left.set_data((X, X), purpose="sample")

    cov_right = EyeCov()
    cov_right.scale = 1.5
    cov_right.set_data((arange(n), arange(n)), purpose="sample")

    cov = SumCov([cov_left, cov_right])

    lik = DeltaProdLik()

    return GGPSampler(lik, mean, cov).sample(random)


def _covariates_sample(random, n, p):
    X = random.randn(n, p)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])
    return X


def test_scan_lmm_iid_prior():
    random = RandomState(9458)
    n = 30
    X = _covariates_sample(random, n, n + 1)
    markers = random.randn(n, 2)

    offset = 1.0

    y = _outcome_sample(random, offset, X)

    lmm = LMM(y, ones((n, 1)), None)

    lmm.fit(verbose=False)
    fast_scanner = lmm.get_fast_scanner()
    lmls = fast_scanner.fast_scan(markers, verbose=False)[0]
    assert_allclose(lmls[:2], [-63.16019973550036, -62.489358539276715])



from __future__ import division

from numpy import exp, log

from optimix import Function, Scalar

from ..util.classes import NamedClass


class GivenCov(NamedClass, Function):
    r"""Given covariance function.

    The mathematical representation is

    .. math::

        f(x_0, x_1) = s \mathrm K_{x_0, x_1},

    where :math:`s` is the scale parameter and :math:`\mathrm K` is given.
    In other words, the user specify passes a covariance matrix, which is
    then indexed by the ``x_0`` and ``x_1``.

    Example
    -------

    .. doctest::

        >>> from glimix_core.cov import GivenCov
        >>> from numpy import dot
        >>> from numpy.random import RandomState
        >>>
        >>> K = RandomState(0).randn(5, 5)
        >>> K = dot(K, K.T)
        >>> cov = GivenCov(K)
        >>> cov.scale = 1.3
        >>> print(cov)
        GivenCov(K=...)
          scale: 1.3
        >>> cov.name = "covname"
        >>> print(cov)
        GivenCov(K=...): covname
          scale: 1.3
    """

    def __init__(self, K):
        Function.__init__(self, logscale=Scalar(0.0))
        self.variables().get("logscale").bounds = (-20.0, +10)
        self._K = K
        NamedClass.__init__(self)

    @property
    def scale(self):
        r"""Scale parameter."""
        return exp(self.variables().get("logscale").value)

    @scale.setter
    def scale(self, scale):
        from numpy_sugar import epsilon

        scale = max(scale, epsilon.tiny)
        self.variables().get("logscale").value = log(scale)

    def value(self, x0, x1):
        r"""Covariance function evaluated at ``(x0, x1)``.

        Parameters
        ----------
        x0 : array_like
            Left-hand side sample indices.
        x1 : array_like
            Right-hand side sample indices.

        Returns
        -------
        array_like
            Submatrix of :math:`s \mathrm K`, row and column-indexed by
            ``x0`` and ``x1``.
        """
        return self.scale * self._K[x0, :][..., x1]

    def gradient(self, x0, x1):
        r"""Derivative of the covariance function evaluated at ``(x0, x1)``.

        Derivative of the covariance function over :math:`\log(s)`.

        Parameters
        ----------
        x0 : array_like
            Left-hand side sample indices.
        x1 : array_like
            Right-hand side sample indices.

        Returns
        -------
        dict
            Dictionary having the `logscale` key for the derivative, row and
            column-indexed by ``x0`` and ``x1``.
        """
        return dict(logscale=self.scale * self._K[x0, :][..., x1])

    def __str__(self):
        tname = type(self).__name__
        msg = "{}(K=...)".format(tname)
        if self.name is not None:
            msg += ": {}".format(self.name)
        msg += "\n"
        msg += "  scale: {}".format(self.scale)
        return msg



from __future__ import division

from numpy import log, pi
from numpy.linalg import slogdet, solve

from optimix import FunctionReduce


class GP(FunctionReduce):
    r"""Gaussian Process inference via maximum likelihood.

    Parameters
    ----------
    y : array_like
        Outcome variable.
    mean : function
        Mean function. (Refer to :doc:`mean`.)
    cov : function
        Covariance function. (Refer to :doc:`cov`.)

    Example
    -------

    .. doctest::

        >>> from numpy.random import RandomState
        >>>
        >>> from glimix_core.example import offset_mean
        >>> from glimix_core.example import linear_eye_cov
        >>> from glimix_core.gp import GP
        >>> from glimix_core.random import GPSampler
        >>>
        >>> random = RandomState(94584)
        >>>
        >>> mean = offset_mean()
        >>> cov = linear_eye_cov()
        >>>
        >>> y = GPSampler(mean, cov).sample(random)
        >>>
        >>> gp = GP(y, mean, cov)
        >>> print('Before: %.4f' % gp.lml())
        Before: -15.5582
        >>> gp.fit(verbose=False)
        >>> print('After: %.4f' % gp.lml())
        After: -13.4791
        >>> print(gp)  # doctest: +FLOAT_CMP
        GP(...)
          lml: -13.47907874997517
          OffsetMean()
            offset: 0.775580366877228
          SumCov(covariances=...)
            LinearCov()
              scale: 2.061153622438558e-09
            EyeCov()
              scale: 0.8675680523425118
    """

    def __init__(self, y, mean, cov):
        from numpy_sugar import is_all_finite

        super(GP, self).__init__([mean, cov], name="GP")

        if not is_all_finite(y):
            raise ValueError("There are non-finite values in the phenotype.")

        self._y = y
        self._cov = cov
        self._mean = mean

    def fit(self, verbose=True, factr=1e5, pgtol=1e-7):
        r"""Maximise the marginal likelihood.

        Parameters
        ----------
        verbose : bool
            ``True`` for progress output; ``False`` otherwise.
            Defaults to ``True``.
        factr : float, optional
            The iteration stops when
            ``(f^k - f^{k+1})/max{|f^k|,|f^{k+1}|,1} <= factr * eps``, where ``eps`` is
            the machine precision.
        pgtol : float, optional
            The iteration will stop when ``max{|proj g_i | i = 1, ..., n} <= pgtol``
            where ``pg_i`` is the i-th component of the projected gradient.

        Notes
        -----
        Please, refer to :func:`scipy.optimize.fmin_l_bfgs_b` for further information
        about ``factr`` and ``pgtol``.
        """
        self.feed().maximize(verbose=verbose, factr=factr, pgtol=pgtol)

    def lml(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            :math:`\log p(\mathbf y)`
        """
        return self.feed().value()

    def _lml_gradient_mean(self, mean, cov, gmean):
        Kiym = solve(cov, self._y - mean)
        return gmean.T.dot(Kiym)

    def _lml_gradient_cov(self, mean, cov, gcov):
        Kiym = solve(cov, self._y - mean)
        return (-solve(cov, gcov).diagonal().sum() + Kiym.dot(gcov.dot(Kiym))) / 2

    def value_reduce(self, values):
        mean = values["GP[0]"]
        cov = values["GP[1]"]
        ym = self._y - mean
        Kiym = solve(cov, ym)

        (s, logdet) = slogdet(cov)
        if not s == 1.0:
            raise RuntimeError("This determinant should not be negative.")

        n = len(self._y)
        return -(logdet + ym.dot(Kiym) + n * log(2 * pi)) / 2

    def gradient_reduce(self, values, gradients):
        mean = values["GP[0]"]
        cov = values["GP[1]"]
        gmean = gradients["GP[0]"]
        gcov = gradients["GP[1]"]

        grad = dict()
        for n, g in iter(gmean.items()):
            grad["GP[0]." + n] = self._lml_gradient_mean(mean, cov, g)

        for n, g in iter(gcov.items()):
            grad["GP[1]." + n] = self._lml_gradient_cov(mean, cov, g)

        return grad

    def __str__(self):
        tname = type(self).__name__
        msg = "{}(...)\n".format(tname)
        msg += "  lml: {}\n".format(self.lml())

        mmsg = str(self._mean).split("\n")
        mmsg = "\n".join(["  " + m for m in mmsg])

        cmsg = str(self._cov).split("\n")
        cmsg = "\n".join(["  " + m for m in cmsg])

        return msg + mmsg + "\n" + cmsg



from __future__ import absolute_import, division, unicode_literals

import warnings

from liknorm import LikNormMachine
from numpy import ascontiguousarray, sign
from numpy.linalg import LinAlgError

from optimix import FunctionReduce

from ..ep import EP
from ..util import check_outcome


class ExpFamGP(FunctionReduce):
    r"""Expectation Propagation for Generalised Gaussian Processes.

    Parameters
    ----------
    y : array_like
        Outcome variable.
    lik_name : str
        Likelihood name.
    mean : function
        Mean function. (Refer to :doc:`mean`.)
    cov : function
        Covariance function. (Refer to :doc:`cov`.)

    Example
    -------

    .. doctest::

        >>> from numpy.random import RandomState
        >>>
        >>> from glimix_core.example import offset_mean
        >>> from glimix_core.example import linear_eye_cov
        >>> from glimix_core.ggp import ExpFamGP
        >>> from glimix_core.lik import BernoulliProdLik
        >>> from glimix_core.link import LogitLink
        >>> from glimix_core.random import GGPSampler
        >>>
        >>> random = RandomState(1)
        >>>
        >>> lik = BernoulliProdLik(LogitLink())
        >>> mean = offset_mean()
        >>> cov = linear_eye_cov()
        >>>
        >>> y = GGPSampler(lik, mean, cov).sample(random)
        >>>
        >>> ggp = ExpFamGP(y, 'bernoulli', mean, cov)
        >>> print('Before: %.4f' % ggp.lml())
        Before: -6.9802
        >>> ggp.fit(verbose=False)
        >>> print('After: %.2f' % ggp.lml())
        After: -6.55
    """

    def __init__(self, y, lik, mean, cov):
        if isinstance(y, tuple):
            n = len(y[0])
        else:
            n = len(y)
        FunctionReduce.__init__(self, [mean, cov], name="ExpFamGP")

        if not isinstance(lik, (tuple, list)):
            lik = (lik,)

        self._lik = (lik[0].lower(),) + tuple(ascontiguousarray(i) for i in lik[1:])
        self._y = check_outcome(y, self._lik)

        self._mean = mean
        self._cov = cov
        self._ep = EP(n)
        self._ep.set_compute_moments(self.compute_moments)

        self._mean_value = None
        self._cov_value = None

        self._machine = LikNormMachine(lik[0], 500)

    def __del__(self):
        if hasattr(self, "_machine"):
            self._machine.finish()

    def fit(self, verbose=True, factr=1e5, pgtol=1e-7):
        r"""Maximise the marginal likelihood.

        Parameters
        ----------
        verbose : bool
            ``True`` for progress output; ``False`` otherwise.
            Defaults to ``True``.
        factr : float, optional
            The iteration stops when
            ``(f^k - f^{k+1})/max{|f^k|,|f^{k+1}|,1} <= factr * eps``, where ``eps`` is
            the machine precision.
        pgtol : float, optional
            The iteration will stop when ``max{|proj g_i | i = 1, ..., n} <= pgtol``
            where ``pg_i`` is the i-th component of the projected gradient.

        Notes
        -----
        Please, refer to :func:`scipy.optimize.fmin_l_bfgs_b` for further information
        about ``factr`` and ``pgtol``.
        """
        self.feed().maximize(verbose=verbose, factr=factr, pgtol=pgtol)

    def lml(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            :math:`\log p(\mathbf y)`
        """
        return self.feed().value()

    def compute_moments(self, eta, tau, moments):
        y = (self._y,) + self._lik[1:]
        self._machine.moments(y, eta, tau, moments)

    def value_reduce(self, values):
        from numpy_sugar import epsilon
        from numpy_sugar.linalg import economic_qs

        mean = values["ExpFamGP[0]"]
        cov = values["ExpFamGP[1]"]
        try:
            self._ep.set_prior(mean, dict(QS=economic_qs(cov)))
            lml = self._ep.lml()
        except (ValueError, LinAlgError) as e:
            warnings.warn(str(e), RuntimeWarning)
            lml = -1 / epsilon.small
        return lml

    def gradient_reduce(self, values, gradients):
        from numpy_sugar import epsilon
        from numpy_sugar.linalg import economic_qs

        mean = values["ExpFamGP[0]"]
        cov = values["ExpFamGP[1]"]
        gmean = gradients["ExpFamGP[0]"]
        gcov = gradients["ExpFamGP[1]"]

        try:
            self._ep.set_prior(mean, dict(QS=economic_qs(cov)))

            grad = dict()

            for n, g in iter(gmean.items()):
                grad["ExpFamGP[0]." + n] = self._ep.lml_derivative_over_mean(g)

            for n, g in iter(gcov.items()):
                QS = economic_qs(g)
                grad["ExpFamGP[1]." + n] = self._ep.lml_derivative_over_cov(QS)

            return grad
        except (ValueError, LinAlgError) as e:
            warnings.warn(str(e), RuntimeWarning)
            v = self.variables().select(fixed=False)
            return {i: -sign(v.get(i).value) / epsilon.small for i in v}



from __future__ import division

from numpy import (
    all as npall,
    asarray,
    atleast_2d,
    clip,
    dot,
    errstate,
    exp,
    full,
    isfinite,
    log,
    maximum,
    sqrt,
    sum as npsum,
    zeros,
)

from glimix_core.util import log2pi
from optimix import Function, Scalar

from ..util import economic_qs_zeros, numbers


class LMMCore(Function):
    def __init__(self, y, X=None, QS=None, SVD=None):
        Function.__init__(self, logistic=Scalar(0.0))
        y = asarray(y, float).ravel()

        if X is not None:
            X = atleast_2d(asarray(X, float).T).T
            if not npall(isfinite(X)):
                msg = "Not all values are finite in the covariates matrix."
                raise ValueError(msg)
            n = X.shape[0]
        else:
            n = SVD[0].shape[0]

        if not npall(isfinite(y)):
            raise ValueError("Not all values are finite in the outcome array.")

        if QS is None:
            QS = economic_qs_zeros(n)
            self.delta = 1.0
            super(LMMCore, self).fix("logistic")
        else:
            self.delta = 0.5

        if not isinstance(QS, tuple):
            raise ValueError(
                "I was expecting a tuple for the covariance decomposition."
            )

        if QS[0][0].shape[0] != y.shape[0]:
            raise ValueError(
                "Number of samples differs between outcome"
                " and covariance decomposition."
            )

        if y.shape[0] != n:
            raise ValueError(
                "Number of samples differs between outcome and covariates."
            )

        self.variables().get("logistic").bounds = (-numbers.logmax, +numbers.logmax)

        self._QS = QS
        self._y = y
        self._scale = 1.0
        self._fix_scale = False
        self._fix_beta = False

        self._tM = None
        self._tbeta = None

        self._svd = None
        self._set_X(X=X, SVD=SVD)
        self._verbose = False

    def _set_X(self, X=None, SVD=None):
        from numpy_sugar.linalg import ddot, economic_svd

        if SVD is None:
            self._svd = economic_svd(X)
        else:
            self._svd = SVD
        self._tM = ddot(self._svd[0], sqrt(self._svd[1]))
        self._tbeta = zeros(self._tM.shape[1])

    def lml(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            Log of the marginal likelihood.
        """
        if self.isfixed("scale"):
            return self._lml_arbitrary_scale()
        return self._lml_optimal_scale()

    def _lml_optimal_scale(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            Log of the marginal likelihood.

        """
        self._update_fixed_effects()

        n = len(self._y)
        lml = -n * log2pi - n - n * log(self.scale)
        lml -= sum(npsum(log(D)) for D in self._D)

        return lml / 2

    def _lml_arbitrary_scale(self):
        r"""Log of the marginal likelihood.

        Returns
        -------
        float
            Log of the marginal likelihood.
        """
        self._update_fixed_effects()
        s = self.scale

        n = len(self._y)
        lml = -n * log2pi - n * log(s)

        lml -= sum(npsum(log(D)) for D in self._D)

        d = (mTQ - yTQ for (mTQ, yTQ) in zip(self._mTQ, self._yTQ))
        lml += sum(dot(j / i, l) for (i, j, l) in zip(self._D, d, self._yTQ)) / s

        return lml / 2

    def _optimal_scale(self):
        from numpy_sugar import epsilon

        if not self.isfixed("beta"):
            return self._optimal_scale_using_optimal_beta()

        yTQDiQTy = self._yTQDiQTy
        yTQDiQTm = self._yTQDiQTm
        b = self._tbeta
        p0 = sum(i - 2 * dot(j, b) for (i, j) in zip(yTQDiQTy, yTQDiQTm))
        p1 = sum(dot(dot(b, i), b) for i in self._mTQDiQTm)
        return maximum((p0 + p1) / len(self._y), epsilon.small)

    def _optimal_scale_using_optimal_beta(self):
        from numpy_sugar import epsilon

        yTQDiQTy = self._yTQDiQTy
        yTQDiQTm = self._yTQDiQTm
        s = sum(i - dot(j, self._tbeta) for (i, j) in zip(yTQDiQTy, yTQDiQTm))
        return maximum(s / len(self._y), epsilon.small)

    def _update_fixed_effects(self):
        from numpy_sugar.linalg import rsolve

        if self.isfixed("beta"):
            return
        yTQDiQTm = list(self._yTQDiQTm)
        mTQDiQTm = list(self._mTQDiQTm)
        nominator = yTQDiQTm[0]
        denominator = mTQDiQTm[0]

        if len(yTQDiQTm) > 1:
            nominator += yTQDiQTm[1]
            denominator += mTQDiQTm[1]

        self._tbeta[:] = rsolve(denominator, nominator)

    @property
    def _D(self):
        D = []
        n = self._y.shape[0]
        if self._QS[1].size > 0:
            D += [self._QS[1] * (1 - self.delta) + self.delta]
        if self._QS[1].size < n:
            D += [full(n - self._QS[1].size, self.delta)]
        return D

    @property
    def _mTQDiQTm(self):
        return (dot(i / j, i.T) for (i, j) in zip(self._tMTQ, self._D))

    @property
    def _mTQ(self):
        return (dot(self.mean.T, Q) for Q in self._QS[0] if Q.size > 0)

    @property
    def _tMTQ(self):
        return (self._tM.T.dot(Q) for Q in self._QS[0] if Q.size > 0)

    @property
    def _yTQ(self):
        return (dot(self._y.T, Q) for Q in self._QS[0] if Q.size > 0)

    @property
    def _yTQQTy(self):
        return (yTQ ** 2 for yTQ in self._yTQ)

    @property
    def _yTQDiQTy(self):
        return (npsum(i / j) for (i, j) in zip(self._yTQQTy, self._D))

    @property
    def _yTQDiQTm(self):
        yTQ = self._yTQ
        D = self._D
        tMTQ = self._tMTQ
        return (dot(i / j, l.T) for (i, j, l) in zip(yTQ, D, tMTQ))

    @property
    def X(self):
        r"""Covariates set by the user.

        It has to be a matrix of number-of-samples by number-of-covariates.

        Returns
        -------
        :class:`numpy.ndarray`
            Covariates.
        """
        from numpy_sugar.linalg import ddot

        return dot(self._svd[0], ddot(self._svd[1], self._svd[2], left=True))

    @X.setter
    def X(self, X):
        self._set_X(X)

    @property
    def beta(self):
        r"""Fixed-effect sizes.

        The optimal fixed-effect sizes is given by any solution to equation

        .. math::

            (\mathrm Q^{\intercal}\mathrm X)^{\intercal}
                \mathrm D^{-1}
                (\mathrm Q^{\intercal}\mathrm X)
                \boldsymbol\beta =
                (\mathrm Q^{\intercal}\mathrm X)^{\intercal}
                \mathrm D^{-1}
                (\mathrm Q^{\intercal}\mathbf y).

        Returns
        -------
        :class:`numpy.ndarray`
            Optimal fixed-effect sizes.
        """
        from numpy_sugar.linalg import ddot, rsolve

        SVs = ddot(self._svd[0], sqrt(self._svd[1]))
        z = rsolve(SVs, self.mean)
        VsD = ddot(sqrt(self._svd[1]), self._svd[2])
        return rsolve(VsD, z)

    @beta.setter
    def beta(self, beta):
        from numpy_sugar.linalg import ddot

        beta = asarray(beta, float)
        VsD = ddot(sqrt(self._svd[1]), self._svd[2])
        self._tbeta[:] = dot(VsD, beta)

    @property
    def delta(self):
        r"""Variance ratio between ``K`` and ``I``."""
        from numpy_sugar import epsilon

        v = float(self.variables().get("logistic").value)
        with errstate(over="ignore", under="ignore"):
            v = 1 / (1 + exp(-v))
        return clip(v, epsilon.tiny, 1 - epsilon.tiny)

    @delta.setter
    def delta(self, delta):
        from numpy_sugar import epsilon

        delta = clip(delta, epsilon.tiny, 1 - epsilon.tiny)
        self.variables().set(dict(logistic=log(delta / (1 - delta))))

    @property
    def scale(self):
        r"""Scaling factor.

        Returns
        -------
        float
            Current scale if fixed; optimal scale otherwise.
        """
        if self.isfixed("scale"):
            return self._scale
        return self._optimal_scale()

    @scale.setter
    def scale(self, scale):
        self._scale = scale

    @property
    def mean(self):
        r"""Mean of the prior.

        Formally, :math:`\mathbf m = \mathrm X \boldsymbol\beta`.

        Returns
        -------
        array_like
            Mean of the prior.
        """
        return dot(self._tM, self._tbeta)

    def mean_star(self, Xstar):
        return dot(Xstar, self.beta)

    def variance_star(self, kss):
        return kss * self.v0 + self.v1

    def covariance_star(self, ks):
        return ks * self.v0

    def covariance(self):
        r"""Covariance of the prior.

        Returns
        -------
        :class:`numpy.ndarray`
            :math:`v_0 \mathrm K + v_1 \mathrm I`.
        """
        from numpy_sugar.linalg import ddot, sum2diag

        Q0 = self._QS[0][0]
        S0 = self._QS[1]
        return sum2diag(dot(ddot(Q0, self.v0 * S0), Q0.T), self.v1)



"""Mean functions.

LinearMean  Dot-product mean function.
OffsetMean  Offset mean.
SumMean     Sum of multiple mean functions.
"""
from .linear import LinearMean
from .offset import OffsetMean
from .sum import SumMean

__all__ = ["OffsetMean", "LinearMean", "SumMean"]



from __future__ import division

from numpy import exp, log, stack

from optimix import Function, Scalar

from ..util.classes import NamedClass


class LinearCov(NamedClass, Function):
    r"""Linear covariance function.

    The mathematical representation is

    .. math::

        f(\mathrm x_0, \mathrm x_1) = s \mathrm x_0^\intercal \mathrm x_1,

    where :math:`s` is the scale parameter.
    """

    def __init__(self):
        Function.__init__(self, logscale=Scalar(0.0))
        self.variables().get("logscale").bounds = (-20.0, +10)
        NamedClass.__init__(self)

    @property
    def scale(self):
        r"""Scale parameter."""
        return exp(self.variables().get("logscale").value)

    @scale.setter
    def scale(self, scale):
        from numpy_sugar import epsilon

        scale = max(scale, epsilon.tiny)
        self.variables().get("logscale").value = log(scale)

    def value(self, x0, x1):
        r"""Covariance function evaluated at ``(x0, x1)``.

        Parameters
        ----------
        x0 : array_like
            Left-hand side sample or samples.
        x1 : array_like
            Right-hand side sample or samples.

        Returns
        -------
        array_like
            :math:`s \mathrm x_0^\intercal \mathrm x_1`.
        """
        x0 = stack(x0, axis=0)
        x1 = stack(x1, axis=0)
        return self.scale * x0.dot(x1.T)

    def gradient(self, x0, x1):
        r"""Derivative of the covariance function evaluated at ``(x0, x1)``.

        Derivative of the covariance function over :math:`\log(s)`:

        .. math::

            s \mathrm x_0^\intercal \mathrm x_1.

        Parameters
        ----------
        x0 : array_like
            Left-hand side sample or samples.
        x1 : array_like
            Right-hand side sample or samples.

        Returns
        -------
        dict
            Dictionary having the `logscale` key for the derivative.
        """
        x0 = stack(x0, axis=0)
        x1 = stack(x1, axis=0)
        return dict(logscale=self.scale * x0.dot(x1.T))

    def __str__(self):
        tname = type(self).__name__
        msg = "{}()".format(tname)
        if self.name is not None:
            msg += ": {}".format(self.name)
        msg += "\n"
        msg += "  scale: {}".format(self.scale)
        return msg



from __future__ import division

from numpy import add

from optimix import FunctionReduce

from ..util.classes import NamedClass


class SumCov(NamedClass, FunctionReduce):
    r"""Sum covariance function.

    The mathematical representation is

    .. math::

        f(f_0, f_1, \dots) = f_0 + f_1 + \dots

    Example
    -------

    .. doctest::

        >>> from glimix_core.cov import LinearCov, SumCov
        >>> from numpy.random import RandomState
        >>>
        >>> random = RandomState(0)
        >>> cov_left = LinearCov()
        >>> cov_right = LinearCov()
        >>>
        >>> cov_left.scale = 0.5
        >>> X0 = random.randn(4, 20)
        >>> X1 = random.randn(4, 15)
        >>>
        >>> cov_left.set_data((X0, X0))
        >>> cov_right.set_data((X1, X1))
        >>>
        >>> cov = SumCov([cov_left, cov_right])
        >>> print(cov)
        SumCov(covariances=...)
          LinearCov()
            scale: 0.5
          LinearCov()
            scale: 1.0
        >>> cov_left.name = "A"
        >>> cov_right.name = "B"
        >>> cov.name = "A+B"
        >>> print(cov)
        SumCov(covariances=...): A+B
          LinearCov(): A
            scale: 0.5
          LinearCov(): B
            scale: 1.0
    """

    def __init__(self, covariances):
        self._covariances = [c for c in covariances]
        FunctionReduce.__init__(self, self._covariances, "sum")
        NamedClass.__init__(self)

    def value_reduce(self, values):
        r"""Sum covariance function evaluated at ``(f_0, f_1, ...)``.

        Parameters
        ----------
        values : dict
            A value for each function involved in the summation.

        Returns
        -------
        dict
            :math:`f_0 + f_1 + \dots`
        """
        return add.reduce(list(values.values()))

    def gradient_reduce(self, values, gradients):
        r"""Sum of covariance function derivatives.

        Parameters
        ----------
        values : dict
            Its value is not used in this particular function. We suggest you to simply
            pass ``None``.
        gradients : dict
            A gradient for each function involved in the summation.

        Returns
        -------
        dict
            :math:`f_0' + f_1' + \dots`
        """
        grad = dict()
        for (gn, gv) in iter(gradients.items()):
            for n, v in iter(gv.items()):
                grad[gn + "." + n] = v
        return grad

    def __str__(self):
        tname = type(self).__name__
        msg = "{}(covariances=...)".format(tname)
        if self.name is not None:
            msg += ": {}".format(self.name)
        for c in self._covariances:
            spl = str(c).split("\n")
            msg = msg + "\n" + "\n".join(["  " + s for s in spl])
        return msg



from __future__ import division

import warnings

from numpy import (
    all as npall,
    asarray,
    atleast_2d,
    clip,
    copyto,
    dot,
    empty,
    full,
    inf,
    isfinite,
    log,
    min as npmin,
    newaxis,
    sum as npsum,
    zeros,
)
from numpy.linalg import LinAlgError

from ..util import hsolve, log2pi


class FastScanner(object):
    r"""Approximated fast inference over several covariates.

    Specifically, it computes the log of the marginal likelihood

    .. math::

        \log p(\mathbf y)_j = \log \mathcal N\big(~ \mathrm X\mathbf b_j^*
        + \mathbf{m}_j \alpha_j^*,~
        s_j^* (\mathrm K + v \mathrm I) ~\big),

    for optimal :math:`\mathbf b_j`, :math:`\alpha_j`, and :math:`s_j^*`
    values.
    Vector :math:`\mathbf{m}_j` is the candidate defined by the i-th column
    of matrix ``M`` provided to method
    :func:`.fast_scan`. Variance :math:`v` is not optimised for performance reasons.
    The method assumes the user has provided a reasonable value for it.

    Parameters
    ----------
    y : array_like
        Real-valued outcome.
    X : array_like
        Matrix of covariates.
    QS : tuple
        Economic eigendecomposition ``((Q0, Q1), S0)`` of ``K``.
    v : float
        Variance due to iid effect.

    Notes
    -----
    The implementation requires further explanation as it is somehow obscure.
    Let :math:`\mathrm B_i=\mathrm Q_i\mathrm D_i^{-1}\mathrm Q_i^{\intercal}`
    for :math:`i \in \{0, 1\}` and
    :math:`\mathrm E_j = [\mathrm X;~ \mathbf m_j]`.
    The matrix resulted from
    :math:`\mathrm E_j^{\intercal}\mathrm B_i\mathrm E_j`
    is represented by the variable ``ETBE``, and
    four views of such a matrix are given by the variables ``XTBX``,
    ``XTBM``, ``MTBX``, and ``MTBM``.
    Those views represent
    :math:`\mathrm X^{\intercal}\mathrm B_i\mathrm X`,
    :math:`\mathrm X^{\intercal}\mathrm B_i\mathbf m_j`,
    :math:`\mathbf m_j^{\intercal}\mathrm B_i\mathrm X`, and
    :math:`\mathbf m_j^{\intercal}\mathrm B_i\mathbf m_j`, respectively.
    """

    def __init__(self, y, X, QS, v):

        y = asarray(y, float)
        X = atleast_2d(asarray(X, float).T).T

        if not npall(isfinite(y)):
            raise ValueError("Not all values are finite in the outcome array.")

        if not npall(isfinite(X)):
            raise ValueError("Not all values are finite in the `X` matrix.")

        if v < 0:
            raise ValueError("Variance has to be non-negative.")

        if not isfinite(v):
            raise ValueError("Variance has to be a finite value..")

        D = []
        if QS[1].size > 0:
            D += [QS[1] + v]
        if QS[1].size < y.shape[0]:
            D += [full(y.shape[0] - QS[1].size, v)]
        yTQ = [dot(y.T, Q) for Q in QS[0] if Q.size > 0]
        XTQ = [dot(X.T, Q) for Q in QS[0] if Q.size > 0]

        yTQDi = [l / r for (l, r) in zip(yTQ, D) if npmin(r) > 0]
        yTBy = npsum([npsum(i * i / j) for (i, j) in zip(yTQ, D) if npmin(j) > 0])
        yTBX = [dot(i, j.T) for (i, j) in zip(yTQDi, XTQ)]
        XTQDi = [i / j for (i, j) in zip(XTQ, D) if npmin(j) > 0]

        self._yTBy = yTBy
        self._yTBX = yTBX
        self._ETBE = [_ETBE(i, j) for (i, j) in zip(XTQDi, XTQ)]
        self._yTBE = [empty(len(i) + 1) for i in yTBX]
        self._XTQ = XTQ
        self._yTQDi = yTQDi
        self._XTQDi = XTQDi
        self._scale = None
        self._QS = QS
        self._D = D
        self._X = X
        self._y = y

    @property
    def _nsamples(self):
        return self._QS[0][0].shape[0]

    @property
    def _ncovariates(self):
        return self._X.shape[1]

    def _static_lml(self):
        n = self._nsamples
        static_lml = -n * log2pi - n
        static_lml -= sum(npsum(_safe_log(D)) for D in self._D)
        return static_lml

    def _fast_scan_chunk(self, M):
        M = asarray(M, float)

        if not M.ndim == 2:
            raise ValueError("`M` array must be bidimensional.")

        if not npall(isfinite(M)):
            raise ValueError("One or more variants have non-finite value.")

        MTQ = [dot(M.T, Q) for Q in self._QS[0] if Q.size > 0]
        yTBM = [dot(i, j.T) for (i, j) in zip(self._yTQDi, MTQ)]
        XTBM = [dot(i, j.T) for (i, j) in zip(self._XTQDi, MTQ)]
        D = self._D
        MTBM = [npsum(i / j * i, 1) for i, j in zip(MTQ, D) if npmin(j) > 0]

        lmls = full(M.shape[1], self._static_lml())
        effsizes = empty(M.shape[1])

        if self._ncovariates == 1:
            self._1covariate_loop(lmls, effsizes, yTBM, XTBM, MTBM)
        else:
            self._multicovariate_loop(lmls, effsizes, yTBM, XTBM, MTBM)

        return lmls, effsizes

    def _multicovariate_loop(self, lmls, effsizes, yTBM, XTBM, MTBM):

        ETBE = self._ETBE
        yTBE = self._yTBE
        tuple_size = len(yTBE)
        if self._scale is not None:
            scale = self._scale

        for i in range(tuple_size):
            yTBE[i][:-1] = self._yTBX[i]

        for i in range(XTBM[0].shape[1]):

            for j in range(tuple_size):
                yTBE[j][-1] = yTBM[j][i]
                ETBE[j].set_XTBM(XTBM[j][:, [i]])
                ETBE[j].set_MTBM(MTBM[j][i])

            beta = _solve(sum(j.value for j in ETBE), sum(j for j in yTBE))
            beta = (beta[:-1][:, newaxis], beta[-1:])
            bstar = _bstar_unpack(beta, self._yTBy, yTBE, ETBE)

            if self._scale is None:
                scale = bstar / self._nsamples
            else:
                lmls[i] += self._nsamples - bstar / scale

            lmls[i] -= self._nsamples * _safe_log(scale)
            effsizes[i] = beta[1][0]

        lmls /= 2

    def _1covariate_loop(self, lmls, effsizes, yTBM, XTBM, MTBM):
        ETBE = self._ETBE
        yTBX = self._yTBX
        XTBX = [i.XTBX for i in ETBE]
        yTBy = self._yTBy

        A00 = sum(i.XTBX[0, 0] for i in ETBE)
        A01 = sum(i[0, :] for i in XTBM)
        A11 = sum(i for i in MTBM)

        b0 = sum(i[0] for i in yTBX)
        b1 = sum(i for i in yTBM)

        beta = hsolve(A00, A01, A11, b0, b1)
        beta = (beta[0][newaxis, :], beta[1])

        bstar = _bstar(beta, yTBy, yTBX, yTBM, XTBX, XTBM, MTBM)

        if self._scale is None:
            scale = bstar / self._nsamples
        else:
            scale = full(len(lmls), self._scale)
            lmls += self._nsamples - bstar / scale

        lmls -= self._nsamples * _safe_log(scale)
        lmls /= 2
        effsizes[:] = beta[1]

    def fast_scan(self, M, verbose=True):
        r"""LML and fixed-effect sizes of each marker.

        If the scaling factor ``s`` is not set by the user via method
        :func:`set_scale`, its optimal value will be found and
        used in the calculation.

        Parameters
        ----------
        M : array_like
            Matrix of fixed-effects across columns.
        verbose : bool, optional
            ``True`` for progress information; ``False`` otherwise.
            Defaults to ``True``.

        Returns
        -------
        :class:`numpy.ndarray`
            Log of the marginal likelihoods.
        :class:`numpy.ndarray`
            Fixed-effect sizes.
        """
        from tqdm import tqdm

        if M.ndim != 2:
            raise ValueError("`M` array must be bidimensional.")
        p = M.shape[1]

        lmls = empty(p)
        effect_sizes = empty(p)

        if verbose:
            nchunks = min(p, 30)
        else:
            nchunks = min(p, 1)

        chunk_size = (p + nchunks - 1) // nchunks

        for i in tqdm(range(nchunks), desc="Scanning", disable=not verbose):
            start = i * chunk_size
            stop = min(start + chunk_size, M.shape[1])

            l, e = self._fast_scan_chunk(M[:, start:stop])

            lmls[start:stop] = l
            effect_sizes[start:stop] = e

        return lmls, effect_sizes

    def null_lml(self):
        r"""Log of the marginal likelihood for the null hypothesis.

        Returns
        -------
        float
            Log of the marginal likelihood.
        """
        n = self._nsamples

        ETBE = self._ETBE
        yTBX = self._yTBX

        A = sum(i.XTBX for i in ETBE)
        b = sum(yTBX)
        beta = _solve(A, b)
        sqrdot = self._yTBy - dot(b, beta)

        lml = self._static_lml()

        if self._scale is None:
            scale = sqrdot / n
        else:
            scale = self._scale
            lml += n
            lml -= sqrdot / scale

        return (lml - n * log(scale)) / 2

    def set_scale(self, scale):
        r"""Set the scaling factor.

        Calling this method disables the automatic scale learning.

        Parameters
        ----------
        scale : float
            Scaling factor.
        """
        self._scale = scale

    def unset_scale(self):
        r"""Unset the scaling factor.

        If called, it enables the scale learning again.
        """
        self._scale = None


class _ETBE(object):
    def __init__(self, XTQDi, XTQ):
        n = XTQDi.shape[0] + 1
        self._data = empty((n, n))
        self._data[:-1, :-1] = dot(XTQDi, XTQ.T)

    @property
    def value(self):
        return self._data

    @property
    def XTBX(self):
        return self._data[:-1, :-1]

    @property
    def XTBM(self):
        return self._data[:-1, -1:]

    @property
    def MTBX(self):
        return self._data[-1:, :-1]

    @property
    def MTBM(self):
        return self._data[-1:, -1]

    def set_XTBM(self, XTBM):
        copyto(self.XTBM, XTBM)
        copyto(self.MTBX, XTBM.T)

    def set_MTBM(self, MTBM):
        copyto(self.MTBM, MTBM)


def _bstar(beta, yTBy, yTBX, yTBM, XTBX, XTBM, MTBM):
    from numpy_sugar.linalg import dotd

    r = full(MTBM[0].shape[0], yTBy)
    r -= 2 * sum(dot(i, beta[0]) for i in yTBX)
    r -= 2 * sum(i * beta[1] for i in yTBM)
    r += sum(dotd(beta[0].T, dot(i, beta[0])) for i in XTBX)
    r += sum(dotd(beta[0].T, i * beta[1]) for i in XTBM)
    r += sum(npsum(beta[1] * i * beta[0], axis=0) for i in XTBM)
    r += sum(beta[1] * i * beta[1] for i in MTBM)
    return r


def _bstar_unpack(beta, yTBy, yTBE, ETBE):
    nc = beta[0].shape[0]
    yTBX = [j[:nc] for j in yTBE]
    yTBM = [j[nc:] for j in yTBE]
    XTBX = [j.XTBX for j in ETBE]
    XTBM = [j.XTBM for j in ETBE]
    MTBM = [j.MTBM for j in ETBE]
    return _bstar(beta, yTBy, yTBX, yTBM, XTBX, XTBM, MTBM)


def _solve(A, y):
    from numpy_sugar.linalg import rsolve

    try:
        beta = rsolve(A, y)
    except LinAlgError:
        msg = "Could not converge to the optimal"
        msg += " effect-size of one of the candidates."
        msg += " Setting its effect-size to zero."
        warnings.warn(msg, RuntimeWarning)
        beta = zeros(A.shape[0])

    return beta


def _safe_log(x):
    from numpy_sugar import epsilon

    return log(clip(x, epsilon.small, inf))



from __future__ import division

from numpy import ascontiguousarray, dot, zeros

from optimix import Function, Vector

from ..util.classes import NamedClass


class LinearMean(NamedClass, Function):
    r"""Linear mean function.

    The mathematical representation is

    .. math::

        f(\mathbf x) = \mathbf x^\intercal \boldsymbol\alpha

    where :math:`\boldsymbol\alpha` is a vector of effect sizes.

    Parameters
    ----------
    size : int
        Number of effects.

    Example
    -------

    .. doctest::

        >>> from numpy import asarray
        >>> from glimix_core.mean import LinearMean
        >>>
        >>> mean = LinearMean(2)
        >>> mean.effsizes = [1.0, -1.0]
        >>> x = [5.1, 1.0]
        >>> print(mean.value(x))
        4.1
        >>> print(mean.gradient(x))
        {'effsizes': [5.1, 1.0]}
        >>> print(mean)
        LinearMean(size=2)
          effsizes: [ 1. -1.]
        >>> mean.name = "covariates"
        >>> print(mean)
        LinearMean(size=2): covariates
          effsizes: [ 1. -1.]
        >>> X = asarray([[5.1,  1.0],
        ...              [0.3, -9.0]])
        >>> mean.set_data(X)
        >>> print(mean.feed().value())
        [4.1 9.3]
        >>> X = asarray([[5.1,  1.0],
        ...              [0.3, -9.0],
        ...              [4.9,  0.0]])
        >>> mean.set_data(X, "dataB")
        >>> print(mean.feed("dataB").value())
        [4.1 9.3 4.9]
    """

    def __init__(self, size):
        Function.__init__(self, effsizes=Vector(zeros(size)))
        self.variables().get("effsizes").bounds = [(-200.0, +200)] * size
        NamedClass.__init__(self)

    def value(self, x):
        r"""Linear mean function.

        Parameters
        ----------
        x : array_like
            Covariates.

        Returns
        -------
        float
            :math:`\mathbf x^\intercal \boldsymbol\alpha`.
        """
        return dot(x, self.variables().get("effsizes").value)

    def gradient(self, x):
        r"""Gradient of the linear mean function.

        Parameters
        ----------
        x : array_like
            Covariates.

        Returns
        -------
        dict
            Dictionary having the `effsizes` key for :math:`\mathbf x`.
        """
        return dict(effsizes=x)

    @property
    def effsizes(self):
        r"""Effect-sizes parameter."""
        return self.variables().get("effsizes").value

    @effsizes.setter
    def effsizes(self, v):
        self.variables().get("effsizes").value = ascontiguousarray(v)

    def __str__(self):
        tname = type(self).__name__
        msg = "{}(size={})".format(tname, len(self.effsizes))
        if self.name is not None:
            msg += ": {}".format(self.name)
        msg += "\n"
        msg += "  effsizes: {}".format(self.effsizes)
        return msg



from __future__ import division

from numpy import arange
from numpy.random import RandomState
from numpy.testing import assert_allclose, assert_array_less, assert_equal

from glimix_core.cov import EyeCov, LinearCov, SumCov
from glimix_core.lik import BinomialProdLik, PoissonProdLik
from glimix_core.link import LogitLink, LogLink
from glimix_core.mean import OffsetMean
from glimix_core.random import (
    GGPSampler,
    GPSampler,
    bernoulli_sample,
    binomial_sample,
    poisson_sample,
)


def test_binomial_sampler():
    random = RandomState(4503)
    link = LogitLink()
    binom = BinomialProdLik(12, link)
    assert_equal(binom.sample(0, random), 7)


def test_poisson_sampler():
    random = RandomState(4503)
    link = LogLink()
    poisson = PoissonProdLik(link)
    assert_equal(poisson.sample(0, random), 1)
    assert_equal(poisson.sample(0, random), 0)
    assert_equal(poisson.sample(0, random), 2)
    assert_equal(poisson.sample(0, random), 0)
    assert_equal(poisson.sample(-10, random), 0)
    assert_equal(poisson.sample(+5, random), 158)


def test_GGPSampler_poisson():
    random = RandomState(4503)
    X = random.randn(10, 15)
    link = LogLink()
    lik = PoissonProdLik(link)

    mean = OffsetMean()
    mean.offset = 1.2
    mean.set_data(arange(10), "sample")
    cov = LinearCov()
    cov.set_data((X, X), "sample")
    sampler = GGPSampler(lik, mean, cov)
    assert_equal(sampler.sample(random), [0, 289, 0, 11, 0, 0, 176, 0, 228, 82])

    mean = OffsetMean()
    mean.offset = 0.0
    mean.set_data(arange(10), "sample")

    cov1 = LinearCov()
    cov1.set_data((X, X), "sample")

    cov2 = EyeCov()
    a = arange(10)
    cov2.set_data((a, a), "sample")

    cov1.scale = 1e-4
    cov2.scale = 1e-4

    cov = SumCov([cov1, cov2])

    sampler = GGPSampler(lik, mean, cov)

    assert_equal(sampler.sample(random), [2, 0, 1, 2, 1, 1, 1, 2, 0, 0])

    cov2.scale = 20.0
    sampler = GGPSampler(lik, mean, cov)
    assert_equal(sampler.sample(random), [0, 0, 0, 2, 0, 0, 1, 22, 0, 0])

    sampler.sample()


def test_GGPSampler_binomial():
    random = RandomState(4503)
    X = random.randn(10, 15)
    link = LogitLink()
    lik = BinomialProdLik(5, link)

    mean = OffsetMean()
    mean.offset = 1.2
    mean.set_data(arange(10), "sample")
    cov = LinearCov()
    cov.set_data((X, X), "sample")
    sampler = GGPSampler(lik, mean, cov)
    assert_equal(sampler.sample(random), [0, 5, 0, 5, 1, 1, 5, 0, 5, 5])

    mean.offset = 0.0
    assert_equal(sampler.sample(random), [5, 4, 1, 0, 0, 1, 4, 5, 5, 0])

    mean = OffsetMean()
    mean.offset = 0.0
    mean.set_data(arange(10), "sample")

    cov1 = LinearCov()
    cov1.set_data((X, X), "sample")

    cov2 = EyeCov()
    a = arange(10)
    cov2.set_data((a, a), "sample")

    cov1.scale = 1e-4
    cov2.scale = 1e-4

    cov = SumCov([cov1, cov2])

    lik = BinomialProdLik(100, link)
    sampler = GGPSampler(lik, mean, cov)
    assert_equal(sampler.sample(random), [56, 56, 55, 51, 59, 45, 47, 43, 51, 38])

    cov2.scale = 100.0
    sampler = GGPSampler(lik, mean, cov)
    assert_equal(sampler.sample(random), [99, 93, 99, 75, 77, 0, 0, 100, 99, 12])


def test_canonical_bernoulli_sampler():
    random = RandomState(9)
    G = random.randn(10, 5)

    y = bernoulli_sample(0.1, G, random_state=random)
    assert_array_less(y, [2] * 10)


def test_canonical_binomial_sampler():
    random = RandomState(9)
    G = random.randn(10, 5)

    y = binomial_sample(5, 0.1, G, random_state=random)
    assert_array_less(y, [5 + 1] * 10)

    ntrials = [2, 3, 1, 1, 4, 5, 1, 2, 1, 1]
    y = binomial_sample(ntrials, -0.1, G, random_state=random)
    assert_array_less(y, [i + 1 for i in ntrials])

    X = random.randn(len(ntrials))
    y = binomial_sample(ntrials, -0.1, G, causal_variants=X, random_state=random)
    assert_array_less(y, [i + 1 for i in ntrials])

    X = random.randn(len(ntrials), 2)
    y = binomial_sample(ntrials, -0.1, G, causal_variants=X, random_state=random)
    assert_array_less(y, [i + 1 for i in ntrials])


def test_canonical_poisson_sampler():
    random = RandomState(10)
    G = random.randn(10, 5)

    y = poisson_sample(0.1, G, random_state=random)
    assert_array_less(y, [20] * len(y))


def test_GPSampler():
    random = RandomState(4503)
    X = random.randn(10, 15)

    mean = OffsetMean()
    mean.offset = 1.2
    mean.set_data(arange(10), "sample")
    cov = LinearCov()
    cov.set_data((X, X), "sample")
    sampler = GPSampler(mean, cov)

    x = [
        -1.34664027,
        5.68720656,
        -6.01941814,
        2.70860707,
        -1.81883975,
        0.21338045,
        5.16462597,
        -2.23134206,
        5.49661095,
        4.38526077,
    ]
    assert_allclose(sampler.sample(random), x)
    sampler.sample()



from __future__ import division

import pytest
from numpy import arange, inf, nan, zeros
from numpy.random import RandomState
from numpy.testing import assert_allclose

from glimix_core.cov import LinearCov
from glimix_core.gp import GP
from glimix_core.mean import LinearMean, OffsetMean
from optimix import check_grad


def test_gp_value_1():
    random = RandomState(94584)
    N = 50
    X = random.randn(N, 100)
    offset = 0.5

    mean = OffsetMean()
    mean.offset = offset
    mean.fix("offset")
    mean.set_data(arange(N))

    cov = LinearCov()
    cov.scale = 1.0
    cov.set_data((X, X))

    y = random.randn(N)

    gp = GP(y, mean, cov)
    assert_allclose(gp.feed().value(), -153.623791551399108)


def test_gp_value_2():
    random = RandomState(94584)
    N = 50
    X1 = random.randn(N, 3)
    X2 = random.randn(N, 100)

    mean = LinearMean(3)
    mean.set_data(X1)

    cov = LinearCov()
    cov.scale = 1.0
    cov.set_data((X2, X2))

    y = random.randn(N)

    gp = GP(y, mean, cov)
    assert_allclose(gp.feed().value(), -153.091074766)

    mean.effsizes = [3.4, 1.11, -6.1]
    assert_allclose(gp.feed().value(), -178.273116338)


def test_gp_gradient():
    random = RandomState(94584)
    N = 50
    X = random.randn(N, 100)
    offset = 0.5

    mean = OffsetMean()
    mean.offset = offset
    mean.fix("offset")
    mean.set_data(arange(N))

    cov = LinearCov()
    cov.scale = 1.0
    cov.set_data((X, X))

    y = random.randn(N)

    gp = GP(y, mean, cov)

    assert_allclose(check_grad(gp.feed()), 0, atol=1e-5)


def test_gp_maximize():
    random = RandomState(94584)
    N = 50
    X = random.randn(N, 100)
    offset = 0.5

    mean = OffsetMean()
    mean.offset = offset
    mean.fix("offset")
    mean.set_data(arange(N))

    cov = LinearCov()
    cov.scale = 1.0
    cov.set_data((X, X))

    y = random.randn(N)

    gp = GP(y, mean, cov)

    assert_allclose(gp.feed().value(), -153.623791551)
    gp.feed().maximize(verbose=False)
    assert_allclose(gp.feed().value(), -79.8992122415)


def test_lmm_nonfinite_phenotype():
    random = RandomState(94584)
    N = 50
    X = random.randn(N, 100)
    offset = 0.5

    mean = OffsetMean()
    mean.offset = offset
    mean.fix("offset")
    mean.set_data(arange(N))

    cov = LinearCov()
    cov.scale = 1.0
    cov.set_data((X, X))

    y = zeros(N)

    y[0] = nan
    with pytest.raises(ValueError):
        GP(y, mean, cov)

    y[0] = -inf
    with pytest.raises(ValueError):
        GP(y, mean, cov)

    y[0] = +inf
    with pytest.raises(ValueError):
        GP(y, mean, cov)
